from __future__ import annotations

import re
from collections import UserList
from collections.abc import Iterable
from collections.abc import Iterable as IterableType
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar

import pandas as pd
from django.core.exceptions import FieldError
from django.db import models
from django.db.models import F, ForeignKey, ManyToManyField, Q, Subquery
from django.db.models.fields.related import ForeignObjectRel
from lamin_utils import logger
from lamindb_setup import settings as setup_settings
from lamindb_setup.core import deprecated
from lamindb_setup.core._docs import doc_args

from ..errors import DoesNotExist, MultipleResultsFound
from ._is_versioned import IsVersioned
from .can_curate import CanCurate, _inspect, _standardize, _validate
from .query_manager import _lookup, _search
from .sqlrecord import Registry, SQLRecord

if TYPE_CHECKING:
    from lamindb.base.types import ListLike, StrField

    from .sqlrecord import Branch

T = TypeVar("T")


pd.set_option("display.max_columns", 200)


# def format_and_convert_to_local_time(series: pd.Series):
#     tzinfo = datetime.now().astimezone().tzinfo
#     timedelta = tzinfo.utcoffset(datetime.now())  # type: ignore
#     return (series + timedelta).dt.strftime("%Y-%m-%d %H:%M:%S %Z")


def get_keys_from_df(data: list, registry: SQLRecord) -> list[str]:
    if len(data) > 0:
        if isinstance(data[0], dict):
            keys = list(data[0].keys())
        else:
            keys = list(data[0].__dict__.keys())
            if "_state" in keys:
                keys.remove("_state")
    else:
        keys = [
            field.name
            for field in registry._meta.fields
            if not isinstance(field, models.ForeignKey)
        ]
        keys += [
            f"{field.name}_id"
            for field in registry._meta.fields
            if isinstance(field, models.ForeignKey)
        ]
    return keys


def get_default_branch_ids(branch: Branch | None = None) -> list[int]:
    """Return branch IDs to include in default queries.

    By default, queries include records on the main branch (branch_id=1) but exclude trashed (branch_id=-1)
    and archived records (branch_id=0). This matches behavior of familiar tools like GitHub, Slack, and
    email clients.

    If a user switches to another branch via `lamin switch branch`, the main branch will still be included.

    Returns:
        List containing the default branch and current branch if different.
    """
    if branch is None:
        branch_id = setup_settings.branch.id
    else:
        branch_id = branch.id
    branch_ids = [branch_id]
    if branch_id != 1:  # add the main branch by default
        branch_ids.append(1)
    return branch_ids


def one_helper(
    self: QuerySet | SQLRecordList,
    does_not_exist_msg: str | None = None,
    raise_doesnotexist: bool = True,
    not_exists: bool | None = None,
    raise_multipleresultsfound: bool = True,
):
    if not_exists is None:
        if isinstance(self, SQLRecordList):
            not_exists = len(self) == 0
        else:
            not_exists = not self.exists()  # type: ignore
    if not_exists:
        if raise_doesnotexist:
            raise DoesNotExist(does_not_exist_msg)
        else:
            return None
    elif len(self) > 1:
        if raise_multipleresultsfound:
            raise MultipleResultsFound(self)
        else:
            return self[0]
    else:
        return self[0]


def get_backward_compat_filter_kwargs(queryset, expressions):
    from lamindb.models import (
        Artifact,
        Collection,
        Transform,
    )

    if queryset.model in {Collection, Transform}:
        name_mappings = {
            "visibility": "branch_id",
            "_branch_code": "branch_id",
        }
    elif queryset.model is Artifact:
        name_mappings = {
            "visibility": "branch_id",
            "_branch_code": "branch_id",
            "transform": "run__transform",
        }
    else:
        return expressions
    was_list = False
    if isinstance(expressions, list):
        was_list = True
        expressions = {field: True for field in expressions}
    mapped = {}
    for field, value in expressions.items():
        parts = field.split("__")
        if parts[0] in name_mappings:
            new_field = name_mappings[parts[0]] + (
                "__" + "__".join(parts[1:]) if len(parts) > 1 else ""
            )
            mapped[new_field] = value
        else:
            mapped[field] = value
    return list(mapped.keys()) if was_list else mapped


def process_expressions(queryset: QuerySet, queries: tuple, expressions: dict) -> dict:
    def _map_databases(value: Any, key: str, target_db: str) -> tuple[str, Any]:
        if isinstance(value, SQLRecord):
            if value._state.db != target_db:
                logger.warning(
                    f"passing record from database {value._state.db} to query {target_db}, matching on uid '{value.uid}'"
                )
                return f"{key}__uid", value.uid
            return key, value

        if (
            key.endswith("__in")
            and isinstance(value, IterableType)
            and not isinstance(value, str)
        ):
            if any(
                isinstance(v, SQLRecord) and v._state.db != target_db for v in value
            ):
                logger.warning(
                    f"passing records from another database to query {target_db}, matching on uids"
                )
                return key.replace("__in", "__uid__in"), [
                    v.uid if isinstance(v, SQLRecord) else v for v in value
                ]
            return key, value

        return key, value

    branch_fields = {"branch", "branch_id"}
    branch_prefixes = ("branch__", "branch_id__")

    def queries_contain_branch(queries: tuple) -> bool:
        """Check if any Q object in queries references branch or branch_id."""

        def check_q_object(q: Q) -> bool:
            # Q objects store their conditions in q.children
            for child in q.children:
                if isinstance(child, tuple) and len(child) == 2:
                    # Normal condition: (key, value)
                    key = child[0]
                    if key in branch_fields or key.startswith(branch_prefixes):
                        return True
                elif isinstance(child, Q):
                    # Nested Q object
                    if check_q_object(child):
                        return True
            return False

        return any(check_q_object(q) for q in queries if isinstance(q, Q))

    expressions = get_backward_compat_filter_kwargs(
        queryset,
        expressions,
    )
    if issubclass(queryset.model, SQLRecord):
        # branch_id is set to 1 unless expressions contains id, uid or hash
        id_uid_hash = {"id", "uid", "hash", "id__in", "uid__in", "hash__in"}
        if not any(expression in id_uid_hash for expression in expressions):
            expressions_have_branch = False
            for expression in expressions:
                if expression in branch_fields or expression.startswith(
                    branch_prefixes
                ):
                    expressions_have_branch = True
                    break
            if not expressions_have_branch and not queries_contain_branch(queries):
                expressions["branch_id__in"] = get_default_branch_ids()
            else:
                # if branch_id is None, do not apply a filter
                # otherwise, it would mean filtering for NULL values, which doesn't make
                # sense for a non-NULLABLE column
                if "branch_id" in expressions and expressions["branch_id"] is None:
                    expressions.pop("branch_id")
                if "branch" in expressions and expressions["branch"] is None:
                    expressions.pop("branch")

    if queryset._db is not None:
        # only check for database mismatch if there is a defined database on the
        # queryset
        return dict(
            (
                _map_databases(value, key, queryset._db)
                for key, value in expressions.items()
            )
        )
    else:
        return expressions


def get(
    registry_or_queryset: Registry | BasicQuerySet,
    idlike: int | str | None = None,
    **expressions,
) -> SQLRecord:
    if isinstance(registry_or_queryset, BasicQuerySet):
        # not QuerySet but only BasicQuerySet
        assert not isinstance(registry_or_queryset, QuerySet)  # noqa: S101

        qs = registry_or_queryset
        registry = qs.model
    else:
        qs = BasicQuerySet(model=registry_or_queryset)
        registry = registry_or_queryset

    if isinstance(idlike, int):
        return qs.get(id=idlike)
    elif isinstance(idlike, str):
        NAME_FIELD = (
            registry._name_field if hasattr(registry, "_name_field") else "name"
        )
        DOESNOTEXIST_MSG = f"No record found with uid '{idlike}'. Did you forget a keyword as in {registry.__name__}.get({NAME_FIELD}='{idlike}')?"
        # this is the case in which the user passes an under-specified uid
        if issubclass(registry, IsVersioned) and len(idlike) <= registry._len_stem_uid:
            new_qs = qs.filter(uid__startswith=idlike, is_latest=True)
            not_exists = None
            if not new_qs.exists():
                # also try is_latest is False due to nothing found
                new_qs = qs.filter(uid__startswith=idlike, is_latest=False)
            else:
                not_exists = False
            # it doesn't make sense to raise MultipleResultsFound when querying with an
            # underspecified uid
            return one_helper(
                new_qs,
                DOESNOTEXIST_MSG,
                not_exists=not_exists,
                raise_multipleresultsfound=False,
            )
        else:
            qs = qs.filter(uid__startswith=idlike)
            return one_helper(qs, DOESNOTEXIST_MSG)
    else:
        assert idlike is None  # noqa: S101
        expressions = process_expressions(qs, [], expressions)
        # inject is_latest for consistency with idlike
        is_latest_was_not_in_expressions = "is_latest" not in expressions
        if issubclass(registry, IsVersioned) and is_latest_was_not_in_expressions:
            expressions["is_latest"] = True
        try:
            return qs.get(**expressions)
        except registry.DoesNotExist as e:
            # handle the case in which the is_latest injection led to a missed query
            if "is_latest" in expressions and is_latest_was_not_in_expressions:
                expressions.pop("is_latest")
                result = qs.filter(**expressions).order_by("-created_at").first()
                if result is not None:
                    return result
            raise e


class SQLRecordList(UserList, Generic[T]):
    """Is ordered, can't be queried, but has `.to_dataframe()`."""

    def __init__(self, records: Iterable[T]):
        if isinstance(records, list):
            self.data = records  # Direct assignment if already a list, no copy
        else:
            super().__init__(records)  # Let UserList handle the conversion

    def to_dataframe(self) -> pd.DataFrame:
        keys = get_keys_from_df(self.data, self.data[0].__class__)
        values = [record.__dict__ for record in self.data]
        return pd.DataFrame(values, columns=keys)

    @deprecated(new_name="to_dataframe")
    def df(self) -> pd.DataFrame:
        return self.to_dataframe()

    def to_list(
        self, field: str
    ) -> list[str]:  # meaningful to be parallel with to_list() in QuerySet
        return [getattr(record, field) for record in self.data]

    @deprecated(new_name="to_list")
    def list(self, field: str) -> list[str]:
        return self.to_list(field)

    def one(self) -> T:
        """Exactly one result. Throws error if there are more or none."""
        return one_helper(self)

    def save(self) -> SQLRecordList[T]:
        """Save all records to the database."""
        from lamindb.models.save import save

        save(self)
        return self


def get_basic_field_names(
    qs: QuerySet,
    include: list[str],
    features_input: bool | list[str] | str,
) -> list[str]:
    exclude_field_names = ["updated_at"]
    field_names = [
        field.name
        for field in qs.model._meta.fields
        if (
            not isinstance(field, models.ForeignKey)
            and field.name not in exclude_field_names
        )
    ]
    field_names += [
        f"{field.name}_id"
        for field in qs.model._meta.fields
        if isinstance(field, models.ForeignKey)
    ]
    for field_name in [
        "version",
        "is_latest",
        "is_locked",
        "run_id",
        "created_at",
        "created_by_id",
        "updated_at",
        "_aux",
        "_real_key",
        "branch_id",
    ]:
        if field_name in field_names:
            field_names.remove(field_name)
            field_names.append(field_name)
    if field_names[0] != "uid" and "uid" in field_names:
        field_names.remove("uid")
        field_names.insert(0, "uid")
    if (
        include or features_input
    ):  # if there is features_input, reduce fields to just the first 3
        subset_field_names = field_names[:3]
        intersection = set(field_names) & set(include)
        subset_field_names += list(intersection)
        field_names = subset_field_names
    return field_names


def get_feature_annotate_kwargs(
    registry: Registry,
    features: bool | list[str] | str | None,
    qs: QuerySet | None = None,
) -> tuple[dict[str, Any], list[str], QuerySet]:
    from lamindb.models import (
        Artifact,
        Feature,
        Record,
        RecordJson,
    )

    if registry not in {Artifact, Record}:
        raise ValueError(
            f"features=True is only applicable for Artifact and Record, not {registry.__name__}"
        )

    if features == "queryset":
        ids_list = qs.values_list("id", flat=True)
        feature_names = []
        for obj in registry._meta.related_objects:
            if not hasattr(getattr(registry, obj.related_name), "through"):
                continue
            link_model = getattr(registry, obj.related_name).through
            if link_model.__name__ == "Record_parents":
                continue
            links = link_model.filter(
                **{registry.__name__.lower() + "_id__in": ids_list}
            )
            feature_names_for_link_model = links.values_list("feature__name", flat=True)
            feature_names += feature_names_for_link_model
        if registry is Record:
            # this request is not strictly necessary, but it makes the resulting reshaped
            # dataframe consistent
            feature_names += RecordJson.filter(record_id__in=ids_list).values_list(
                "feature__name", flat=True
            )
        features = list(set(feature_names))  # remove duplicates

    feature_qs = Feature.filter()
    if isinstance(features, list):
        feature_qs = feature_qs.filter(name__in=features)
        feature_names = features
    else:  # features is True -- only consider categorical features from ULabel and non-categorical features
        feature_qs = feature_qs.filter(
            Q(~Q(dtype__startswith="cat["))
            | Q(dtype__startswith="cat[ULabel")
            | Q(dtype__startswith="cat[Record")
        )
        feature_names = feature_qs.to_list("name")
        logger.important(
            f"queried for all categorical features with dtype ULabel or Record and non-categorical features: ({len(feature_names)}) {feature_names}"
        )
    # Get the categorical features
    cat_feature_types = {
        feature.dtype.replace("cat[", "").replace("]", "")
        for feature in feature_qs
        if feature.dtype.startswith("cat[")
    }
    # Get relationships of labels and features
    link_models_on_models = {
        getattr(
            registry, obj.related_name
        ).through.__get_name_with_module__(): obj.related_model.__get_name_with_module__()
        for obj in registry._meta.related_objects
        if obj.related_model.__get_name_with_module__() in cat_feature_types
    }
    if registry is Artifact:
        link_models_on_models["ArtifactULabel"] = "ULabel"
    else:
        link_models_on_models["RecordRecord"] = "Record"
    link_attributes_on_models = {
        obj.related_name: link_models_on_models[
            obj.related_model.__get_name_with_module__()
        ]
        for obj in registry._meta.related_objects
        if (
            obj.related_model.__get_name_with_module__() in link_models_on_models
            and (
                not obj.related_name.startswith("links_record")
                if registry is Record
                else True
            )
        )
    }
    # Prepare Django's annotate for features
    annotate_kwargs = {}
    for link_attr, feature_type in link_attributes_on_models.items():
        if link_attr == "links_project" and registry is Record:
            # we're only interested in _values_project when "annotating" records
            continue
        annotate_kwargs[f"{link_attr}__feature__name"] = F(
            f"{link_attr}__feature__name"
        )
        if registry is Artifact:
            field_name = (
                feature_type.split(".")[1] if "." in feature_type else feature_type
            ).lower()
        else:
            field_name = "value"
        annotate_kwargs[f"{link_attr}__{field_name}__name"] = F(
            f"{link_attr}__{field_name}__name"
        )
    json_values_attribute = "_feature_values" if registry is Artifact else "values_json"
    annotate_kwargs[f"{json_values_attribute}__feature__name"] = F(
        f"{json_values_attribute}__feature__name"
    )
    annotate_kwargs[f"{json_values_attribute}__value"] = F(
        f"{json_values_attribute}__value"
    )
    return annotate_kwargs, feature_names, feature_qs


# https://claude.ai/share/16280046-6ae5-4f6a-99ac-dec01813dc3c
def analyze_lookup_cardinality(
    model_class: SQLRecord, lookup_paths: list[str] | None
) -> dict[str, str]:
    """Analyze lookup cardinality.

    Analyzes Django model lookups to determine if they will result in
    one-to-one or one-to-many relationships when used in annotations.

    Args:
        model_class: The Django model class to analyze
        include: List of lookup paths (e.g. ["created_by__name", "ulabels__name"])

    Returns:
        Dictionary mapping lookup paths to either 'one' or 'many'
    """
    result = {}  # type: ignore
    if lookup_paths is None:
        return result
    for lookup_path in lookup_paths:
        parts = lookup_path.split("__")
        current_model = model_class
        is_many = False

        # Walk through each part of the lookup path
        for part in parts[:-1]:  # Exclude the last part as it's an attribute
            field = None

            # Handle reverse relations
            for f in current_model._meta.get_fields():
                if isinstance(f, ForeignObjectRel) and f.get_accessor_name() == part:
                    field = f
                    is_many = not f.one_to_one
                    if hasattr(f, "field"):
                        current_model = f.field.model
                    break

            # Handle forward relations
            if field is None:
                field = current_model._meta.get_field(part)
                if isinstance(field, ManyToManyField):
                    is_many = True
                    current_model = field.remote_field.model
                elif isinstance(field, ForeignKey):
                    current_model = field.remote_field.model

        result[lookup_path] = "many" if is_many else "one"

    return result


def reorder_subset_columns_in_df(
    df: pd.DataFrame, column_order: list[str], position=3
) -> pd.DataFrame:
    valid_columns = [col for col in column_order if col in df.columns]
    all_cols = df.columns.tolist()
    remaining_cols = [col for col in all_cols if col not in valid_columns]
    new_order = remaining_cols[:position] + valid_columns + remaining_cols[position:]
    return df[new_order]


# https://lamin.ai/laminlabs/lamindata/transform/BblTiuKxsb2g0003
# https://claude.ai/chat/6ea2498c-944d-4e7a-af08-29e5ddf637d2
def reshape_annotate_result(
    registry: Registry,
    df: pd.DataFrame,
    field_names: list[str],
    cols_from_include: dict[str, str] | None,
    feature_names: list[str],
    feature_qs: QuerySet | None,
) -> pd.DataFrame:
    """Reshapes tidy table to wide format.

    Args:
        field_names: List of basic fields to include in result
        df: Input dataframe with experimental data
        extra_columns: Dict specifying additional columns to process with types ('one' or 'many')
            e.g., {'ulabels__name': 'many', 'created_by__name': 'one'}
        feature_names: Feature names.
    """
    from lamindb.models import Artifact

    cols_from_include = cols_from_include or {}

    json_values_attribute = "_feature_values" if registry is Artifact else "values_json"

    # initialize result with basic fields, need a copy as we're modifying it
    # will give us warnings otherwise
    result = df[field_names].copy()
    # process features if requested
    if feature_names:
        # handle json values
        feature_cols = [
            f"{json_values_attribute}__feature__name",
            f"{json_values_attribute}__value",
        ]
        if all(col in df.columns for col in feature_cols):
            # Create two separate dataframes - one for dict values and one for non-dict values
            is_dict = df[f"{json_values_attribute}__value"].apply(
                lambda x: isinstance(x, dict)
            )
            dict_df, non_dict_df = df[is_dict], df[~is_dict]

            # Process non-dict values using set aggregation
            non_dict_features = non_dict_df.groupby(
                ["id", f"{json_values_attribute}__feature__name"]
            )[f"{json_values_attribute}__value"].agg(set)

            # Process dict values using first aggregation
            dict_features = dict_df.groupby(
                ["id", f"{json_values_attribute}__feature__name"]
            )[f"{json_values_attribute}__value"].agg("first")

            # Combine the results
            combined_features = pd.concat([non_dict_features, dict_features])

            # Unstack and reset index
            feature_values = combined_features.unstack().reset_index()
            if not feature_values.empty:
                result = result.join(
                    feature_values.set_index("id"),
                    on="id",
                )

        # handle categorical features
        links_prefix = "links_" if registry is Artifact else ("links_", "values_")
        links_features = [
            col
            for col in df.columns
            if "feature__name" in col and col.startswith(links_prefix)
        ]

        if links_features:
            result = process_links_features(df, result, links_features, feature_names)

        def extract_single_element(s):
            if not hasattr(s, "__len__"):  # is NaN or other scalar
                return s
            if len(s) != 1:
                # TODO: below should depend on feature._expect_many
                # logger.warning(
                #     f"expected single value because `feature._expect_many is False` but got set {len(s)} elements: {s}"
                # )
                return s
            return next(iter(s))

        for feature in feature_qs:
            if feature.name in result.columns:
                # TODO: make dependent on feature._expect_many through
                # lambda x: extract_single_element(x, feature)
                result[feature.name] = result[feature.name].apply(
                    extract_single_element
                )
                if feature.dtype.startswith("cat"):
                    try:
                        # Try to convert to category - this will fail if complex objects remain
                        result[feature.name] = result[feature.name].astype("category")
                    except (TypeError, ValueError):
                        # If conversion fails, the column still contains complex objects
                        pass
                if feature.dtype.startswith("datetime"):
                    try:
                        # Try to convert to category - this will fail if complex objects remain
                        result[feature.name] = pd.to_datetime(result[feature.name])
                    except (TypeError, ValueError):
                        # If conversion fails, the column still contains complex objects
                        pass

        # sort columns
        result = reorder_subset_columns_in_df(result, feature_names)

    if cols_from_include:
        result = process_cols_from_include(df, result, cols_from_include)

    return result.drop_duplicates(subset=["id"])


def process_links_features(
    df: pd.DataFrame,
    result: pd.DataFrame,
    feature_cols: list[str],
    features: bool | list[str],
) -> pd.DataFrame:
    """Process links_XXX feature columns."""
    # this loops over different entities that might be linked under a feature
    for feature_col in feature_cols:
        links_attribute = "links_" if feature_col.startswith("links_") else "values_"
        regex = f"{links_attribute}(.+?)__feature__name"
        prefix = re.match(regex, feature_col).group(1)

        value_cols = [
            col
            for col in df.columns
            if col.startswith(f"{links_attribute}{prefix}__")
            and col.endswith("__name")
            and "feature__name" not in col
        ]

        if not value_cols:
            continue

        value_col = value_cols[0]
        feature_names = df[feature_col].unique()
        feature_names = feature_names[~pd.isna(feature_names)]

        # Filter features if specific ones requested
        if isinstance(features, list):
            feature_names = [f for f in feature_names if f in features]

        for feature_name in feature_names:
            mask = df[feature_col] == feature_name
            feature_values = df[mask].groupby("id")[value_col].agg(set)
            result.insert(3, feature_name, result["id"].map(feature_values))

    return result


def process_cols_from_include(
    df: pd.DataFrame, result: pd.DataFrame, extra_columns: dict[str, str]
) -> pd.DataFrame:
    """Process additional columns based on their specified types."""
    for col, col_type in extra_columns.items():
        if col not in df.columns:
            continue
        if col in result.columns:
            continue

        values = df.groupby("id")[col].agg(set if col_type == "many" else "first")
        result.insert(3, col, result["id"].map(values))

    return result


def _queryset_class_factory(
    registry: Registry, queryset_cls: type[models.QuerySet]
) -> type[models.QuerySet]:
    from lamindb.models import Artifact, ArtifactSet

    # If the model is Artifact, create a new class
    # for BasicQuerySet or QuerySet that inherits from ArtifactSet.
    # This allows to add artifact specific functionality to all classes
    # inheriting from BasicQuerySet.
    # Thus all query sets of artifacts (and only of artifacts)
    # will have functions from ArtifactSet.
    if registry is Artifact and not issubclass(queryset_cls, ArtifactSet):
        new_cls = type(
            "Artifact" + queryset_cls.__name__, (queryset_cls, ArtifactSet), {}
        )
    else:
        new_cls = queryset_cls

    return new_cls


class BasicQuerySet(models.QuerySet):
    """Sets of records returned by queries.

    See Also:

        `django QuerySet <https://docs.djangoproject.com/en/stable/ref/models/querysets/>`__

    Examples:

        Any filter statement produces a query set::

            queryset = Registry.filter(name__startswith="keyword")
    """

    def __new__(cls, model=None, query=None, using=None, hints=None):
        # see comments in _queryset_class_factory
        return object.__new__(_queryset_class_factory(model, cls))

    def _to_class(
        self, cls: type[models.QuerySet], copy: bool = True
    ) -> models.QuerySet:
        qs = self.all() if copy else self
        qs.__class__ = cls
        return qs

    def _to_basic(self, copy: bool = True) -> BasicQuerySet:
        cls = _queryset_class_factory(self.model, BasicQuerySet)
        return self._to_class(cls, copy)

    def _to_non_basic(self, copy: bool = True) -> QuerySet:
        cls = _queryset_class_factory(self.model, QuerySet)
        return self._to_class(cls, copy)

    @doc_args(SQLRecord.to_dataframe.__doc__)
    def to_dataframe(
        self,
        include: str | list[str] | None = None,
        features: bool | list[str] | str | None = None,
    ) -> pd.DataFrame:
        """{}"""  # noqa: D415
        time = datetime.now(timezone.utc)
        if include is None:
            include_input = []
        elif isinstance(include, str):
            include_input = [include]
        else:
            include_input = include
        features_input = [] if features is None else features
        include = get_backward_compat_filter_kwargs(self, include_input)
        field_names = get_basic_field_names(self, include_input, features_input)

        annotate_kwargs = {}
        feature_names: list[str] = []
        feature_qs = None
        if features:
            feature_annotate_kwargs, feature_names, feature_qs = (
                get_feature_annotate_kwargs(self.model, features, self)
            )
            time = logger.debug("finished feature_annotate_kwargs", time=time)
            annotate_kwargs.update(feature_annotate_kwargs)
        if include_input:
            include_input = include_input.copy()[::-1]  # type: ignore
            include_kwargs = {s: F(s) for s in include_input if s not in field_names}
            annotate_kwargs.update(include_kwargs)
        if annotate_kwargs:
            id_subquery = self.values("id")
            time = logger.debug("finished get id values", time=time)
            # for annotate, we want the queryset without filters so that joins don't affect the annotations
            query_set_without_filters = self.model.objects.using(self._db).filter(
                id__in=Subquery(id_subquery)
            )
            time = logger.debug("finished get query_set_without_filters", time=time)
            if self.query.order_by:
                # Apply the same ordering to the new queryset
                query_set_without_filters = query_set_without_filters.order_by(
                    *self.query.order_by
                )
                time = logger.debug("finished order by", time=time)
            queryset = query_set_without_filters.annotate(**annotate_kwargs)
            time = logger.debug("finished annotate", time=time)
        else:
            queryset = self

        df = pd.DataFrame(queryset.values(*field_names, *list(annotate_kwargs.keys())))
        if len(df) == 0:
            df = pd.DataFrame({}, columns=field_names)
            return df
        time = logger.debug("finished creating first dataframe", time=time)
        cols_from_include = analyze_lookup_cardinality(self.model, include_input)  # type: ignore
        time = logger.debug("finished analyze_lookup_cardinality", time=time)
        df_reshaped = reshape_annotate_result(
            self.model, df, field_names, cols_from_include, feature_names, feature_qs
        )
        time = logger.debug("finished reshape_annotate_result", time=time)
        pk_name = self.model._meta.pk.name
        pk_column_name = pk_name if pk_name in df.columns else f"{pk_name}_id"
        if pk_column_name in df_reshaped.columns:
            df_reshaped = df_reshaped.set_index(pk_column_name)
        time = logger.debug("finished", time=time)
        return df_reshaped

    @deprecated(new_name="to_dataframe")
    def df(
        self,
        include: str | list[str] | None = None,
        features: bool | list[str] | str | None = None,
    ) -> pd.DataFrame:
        return self.to_dataframe(include, features)

    def delete(self, *args, permanent: bool | None = None, **kwargs):
        """Delete all records in the query set.

        Args:
            permanent: Whether to permanently delete the record (skips trash).
                Is only relevant for records that have the `branch` field.
                If `None`, uses soft delete for records that have the `branch` field,
                hard delete otherwise.

        Note:
            Calling `delete()` twice on the same queryset does NOT permanently delete in bulk operations.
            Use `permanent=True` for actual deletion.

        Examples:

            For any `QuerySet` object `qs`, call:

            >>> qs.delete()
        """
        from lamindb.models import Artifact, Collection, Run, Storage, Transform

        # all these models have non-trivial delete behavior, hence we need to handle in a loop
        if self.model in {Artifact, Collection, Transform, Run}:
            for record in self:
                record.delete(*args, permanent=permanent, **kwargs)
        elif self.model is Storage:  # storage does not have soft delete
            if permanent is False:
                raise ValueError(
                    "Soft delete is not possible for Storage, "
                    "use 'permanent=True' or 'permanent=None' for permanent deletion."
                )
            for record in self:
                record.delete()
        else:
            if not permanent and hasattr(self.model, "branch_id"):
                logger.warning("moved records to trash (branch_id = -1)")
                self.update(branch_id=-1)
            else:
                if permanent is False:
                    raise ValueError(
                        f"Soft delete is not possible for {self.model.__name__}, "
                        "use 'permanent=True' for permanent deletion."
                    )
                super().delete(*args, **kwargs)

    def to_list(self, field: str | None = None) -> list[SQLRecord] | list[str]:
        """Populate an (unordered) list with the results.

        Note that the order in this list is only meaningful if you ordered the underlying query set with `.order_by()`.

        Examples:
            >>> queryset.to_list()  # list of records
            >>> queryset.to_list("name")  # list of values
        """
        if field is None:
            return list(self)
        else:
            # list casting is necessary because values_list does not return a list
            return list(self.values_list(field, flat=True))

    @deprecated(new_name="to_list")
    def list(self, field: str | None = None) -> list[SQLRecord] | list[str]:
        return self.to_list(field)

    def first(self) -> SQLRecord | None:
        """If non-empty, the first result in the query set, otherwise ``None``.

        Examples:
            >>> queryset.first()
        """
        if len(self) == 0:
            return None
        return self[0]

    def one(self) -> SQLRecord:
        """Exactly one result. Raises error if there are more or none."""
        return one_helper(self)

    def one_or_none(self) -> SQLRecord | None:
        """At most one result. Returns it if there is one, otherwise returns ``None``.

        Examples:
            >>> ULabel.filter(name="benchmark").one_or_none()
            >>> ULabel.filter(name="non existing label").one_or_none()
        """
        return one_helper(self, raise_doesnotexist=False)

    def latest_version(self) -> QuerySet:
        """Filter every version family by latest version."""
        if issubclass(self.model, IsVersioned):
            return self.filter(is_latest=True)
        else:
            raise ValueError("SQLRecord isn't subclass of `lamindb.core.IsVersioned`")

    @doc_args(_search.__doc__)
    def search(self, string: str, **kwargs):
        """{}"""  # noqa: D415
        return _search(cls=self, string=string, **kwargs)

    @doc_args(_lookup.__doc__)
    def lookup(self, field: StrField | None = None, **kwargs) -> NamedTuple:
        """{}"""  # noqa: D415
        return _lookup(cls=self, field=field, **kwargs)

    # -------------------------------------------------------------------------------------
    # CanCurate
    # -------------------------------------------------------------------------------------

    @doc_args(CanCurate.validate.__doc__)
    def validate(self, values: ListLike, field: str | StrField | None = None, **kwargs):
        """{}"""  # noqa: D415
        return _validate(cls=self, values=values, field=field, **kwargs)

    @doc_args(CanCurate.inspect.__doc__)
    def inspect(self, values: ListLike, field: str | StrField | None = None, **kwargs):
        """{}"""  # noqa: D415
        return _inspect(cls=self, values=values, field=field, **kwargs)

    @doc_args(CanCurate.standardize.__doc__)
    def standardize(
        self, values: Iterable, field: str | StrField | None = None, **kwargs
    ):
        """{}"""  # noqa: D415
        return _standardize(cls=self, values=values, field=field, **kwargs)


# this differs from BasicQuerySet only in .filter and .get
# QueryManager returns BasicQuerySet because it is problematic to redefine .filter and .get
# for a query set used by the default manager
class QuerySet(BasicQuerySet):
    """Sets of records returned by queries.

    Implements additional filtering capabilities.

    See Also:

        `django QuerySet <https://docs.djangoproject.com/en/4.2/ref/models/querysets/>`__

    Examples:

        >>> ULabel(name="my label").save()
        >>> queryset = ULabel.filter(name="my label")
        >>> queryset # an instance of QuerySet
    """

    def _handle_unknown_field(self, error: FieldError) -> None:
        """Suggest available fields if an unknown field was passed."""
        if "Cannot resolve keyword" in str(error):
            field = str(error).split("'")[1]
            avail_fields = self.model.__get_available_fields__()
            if "_branch_code" in avail_fields:
                avail_fields.remove("_branch_code")  # backward compat
            fields = ", ".join(sorted(avail_fields))
            raise FieldError(
                f"Unknown field '{field}'. Available fields: {fields}"
            ) from None
        raise error  # pragma: no cover

    def get(self, idlike: int | str | None = None, **expressions) -> SQLRecord:
        """Query a single record. Raises error if there are more or none."""
        is_run_input = expressions.pop("is_run_input", False)

        # artifacts_from_path and get accept only BasicQuerySet
        qs = self._to_class(BasicQuerySet, copy=True)

        if path := expressions.pop("path", None):
            from .artifact_set import ArtifactSet, artifacts_from_path

            if not isinstance(self, ArtifactSet):
                raise ValueError("Querying by path is only possible for artifacts.")
            qs = artifacts_from_path(qs, path)

        try:
            record = get(qs, idlike, **expressions)
        except ValueError as e:
            # Pass through original error for explicit id lookups
            if "Field 'id' expected a number" in str(e):
                if "id" in expressions:
                    raise
                field = next(iter(expressions))
                raise FieldError(
                    f"Invalid lookup '{expressions[field]}' for {field}. Did you mean {field}__name?"
                ) from None
            raise  # pragma: no cover
        except FieldError as e:
            self._handle_unknown_field(e)
            raise  # pragma: no cover

        if is_run_input is not False:  # might be None or True or Run
            from .artifact import Artifact, track_run_input
            from .collection import Collection

            if isinstance(record, (Artifact, Collection)):
                track_run_input(record, is_run_input)

        return record

    def filter(self, *queries, **expressions) -> QuerySet:
        """Query a set of records."""
        from lamindb.models import Artifact, Record, Run

        registry = self.model

        if not expressions.pop("_skip_filter_with_features", False) and registry in {
            Artifact,
            Run,
            Record,
        }:
            from ._feature_manager import filter_with_features

            return filter_with_features(self, *queries, **expressions)

        # Suggest to use __name for related fields such as id when not passed
        for field, value in expressions.items():
            if (
                isinstance(value, str)
                and value.strip("-").isalpha()
                and "__" not in field
                and hasattr(registry, field)
            ):
                field_attr = getattr(registry, field)
                if hasattr(field_attr, "field") and field_attr.field.related_model:
                    raise FieldError(
                        f"Invalid lookup '{value}' for {field}. Did you mean {field}__name?"
                    )

        expressions = process_expressions(self, queries, expressions)
        # need to run a query if queries or expressions are not empty
        if queries or expressions:
            try:
                return super().filter(*queries, **expressions)
            except FieldError as e:
                self._handle_unknown_field(e)
        return self



import lamindb_setup as ln_setup
from laminci.db import setup_local_test_postgres

pgurl = setup_local_test_postgres()

ln_setup.init(
    storage="./default_storage_permissions",
    name="lamindb-test-permissions",
    db=pgurl,
)

# can't add this app in the init because don't want t trigger the initial migration
# that conflicts with _install_db_module
ln_setup.settings.instance._schema_str = "hubmodule"
ln_setup.settings.instance._persist()



import lamindb as ln
import pandas as pd
from lamindb.examples.fixtures.sheets import (
    populate_nextflow_sheet_with_samples,  # noqa: F401
    populate_sheets_compound_treatment,  # noqa: F401
)


def test_record_example_compound_treatment(
    populate_sheets_compound_treatment: tuple[ln.Record, ln.Record],  # noqa: F811
):
    treatments_sheet, sample_sheet1 = populate_sheets_compound_treatment

    dictionary = (
        ln.Record.filter(type=treatments_sheet)
        .to_dataframe()[["is_type", "name"]]
        .to_dict(orient="list")
    )
    assert dictionary == {
        "is_type": [
            False,
            False,
        ],
        "name": [
            "treatment1",
            "treatment2",
        ],
    }

    dictionary = (
        ln.Record.filter(type=treatments_sheet)
        .to_dataframe(features=True)[["compound", "concentration", "name"]]
        .to_dict(orient="list")
    )
    assert dictionary == {
        "compound": [
            "drug1",
            "drug2",
        ],
        "concentration": [
            "2nM",
            "4nM",
        ],
        "name": [
            "treatment1",
            "treatment2",
        ],
    }

    dictionary = (
        ln.Record.filter(type=sample_sheet1)
        .to_dataframe(features=["cell_line", "treatment"])[
            ["cell_line", "name", "treatment"]
        ]
        .to_dict(orient="list")
    )
    assert dictionary == {
        "cell_line": [
            "HEK293T cell",
            "HEK293T cell",
        ],
        "name": [
            "sample1",
            "sample2",
        ],
        "treatment": [
            "treatment1",
            "treatment2",
        ],
    }

    df = sample_sheet1.type_to_dataframe()
    dictionary = df[
        ["cell_line", "__lamindb_record_name__", "treatment", "preparation_date"]
    ].to_dict(orient="list")
    assert dictionary == {
        "cell_line": [
            "HEK293T cell",
            "HEK293T cell",
        ],
        "__lamindb_record_name__": [
            "sample1",
            "sample2",
        ],
        "preparation_date": [
            pd.to_datetime("2025-06-01T05:00:00"),
            pd.to_datetime("2025-06-01T06:00:00"),
        ],
        "treatment": [
            "treatment1",
            "treatment2",
        ],
    }

    artifact = sample_sheet1.to_artifact()
    assert sample_sheet1.schema.members.to_list("name") == [
        "treatment",
        "cell_line",
        "preparation_date",
        "project",
    ]
    assert artifact.run.input_records.count() == 1
    assert artifact.transform.type == "function"
    # looks something like this:
    # treatment,cell_line,preparation_date,__lamindb_record_uid__,__lamindb_record_name__
    # treatment1,HEK293T cell,2025-06-01 05:00:00,iCwgKgZELoLtIoGy,sample1
    # treatment2,HEK293T cell,2025-06-01 06:00:00,qvU9m7VF6fSdsqJs,sample2
    assert len(artifact.load()) == 2  # two rows in the dataframe
    assert artifact.path.read_text().startswith("""\
treatment,cell_line,preparation_date,project,__lamindb_record_uid__,__lamindb_record_name__
treatment1,HEK293T cell,2025-06-01 05:00:00,Project 1""")
    assert artifact.key == f"sheet_exports/{sample_sheet1.name}.csv"
    assert artifact.description.startswith(f"Export of sheet {sample_sheet1.uid}")
    assert artifact._state.adding is False
    assert ln.models.ArtifactRecord.filter(artifact=artifact).count() == 2
    assert (
        artifact.features.describe(return_str=True)
        == """\
Artifact .csv · DataFrame · dataset
└── Dataset features
    └── columns • 4         [Feature]
        cell_line           cat[bionty.CellLine]    HEK293T cell
        project             cat[Project]            Project 1
        treatment           cat[Record[Treatment]]  treatment1, treatment2
        preparation_date    datetime"""
    )
    # re-run the export which triggers hash lookup
    sample_sheet1.to_artifact()
    # soft-delete a record in the sheet
    sample_sheet1.records.first().delete()
    assert ln.Record.filter(type=sample_sheet1).count() == 1
    df = sample_sheet1.type_to_dataframe()
    assert len(df) == 1  # one row in the dataframe

    artifact.delete(permanent=True)


def test_nextflow_sheet_with_samples(
    populate_nextflow_sheet_with_samples: ln.Record,  # noqa: F811
):
    """Test the example fixture for nextflow sheet with samples."""
    # This test is to ensure that the fixture works as expected
    # and that the data is correctly populated in the database.
    nextflow_sheet = populate_nextflow_sheet_with_samples

    df = nextflow_sheet.to_pandas()

    assert df[
        ["expected_cells", "fastq_1", "fastq_2", "sample", "__lamindb_record_name__"]
    ].to_dict(orient="list") == {
        "expected_cells": [
            5000,
            5000,
            5000,
        ],
        "fastq_1": [
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_X_S1_L001_R1_001.fastq.gz",
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_Y_S1_L001_R1_001.fastq.gz",
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_Y_S1_L002_R1_001.fastq.gz",
        ],
        "fastq_2": [
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_X_S1_L001_R2_001.fastq.gz",
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_Y_S1_L001_R2_001.fastq.gz",
            "https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_Y_S1_L002_R2_001.fastq.gz",
        ],
        "__lamindb_record_name__": [
            None,
            None,
            None,
        ],
        "sample": [
            "Sample_X",
            "Sample_Y",
            "Sample_Y",
        ],
    }

    assert nextflow_sheet.schema is not None
    artifact = nextflow_sheet.to_artifact()
    assert artifact.schema is nextflow_sheet.schema
    assert artifact._state.adding is False
    assert nextflow_sheet.schema.members.to_list("name") == [
        "sample",
        "fastq_1",
        "fastq_2",
        "expected_cells",
        "seq_center",
    ]
    assert artifact.features.slots["columns"].members.to_list("name") == [
        "sample",
        "fastq_1",
        "fastq_2",
        "expected_cells",
    ]
    print(artifact.path.read_text())
    print(artifact.features.describe(return_str=True))
    assert artifact.path.read_text().startswith("""\
sample,fastq_1,fastq_2,expected_cells,__lamindb_record_uid__,__lamindb_record_name__
Sample_X,https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_X_S1_L001_R1_001.fastq.gz,https://raw.githubusercontent.com/nf-core/test-datasets/scrnaseq/testdata/cellranger/Sample_X_S1_L001_R2_001.fastq.gz,5000,""")
    assert (
        artifact.features.describe(return_str=True)
        == """\
Artifact .csv · DataFrame · dataset
└── Dataset features
    └── columns • 4         [Feature]
        sample              cat[Record[BioSample]]  Sample_X, Sample_Y
        fastq_1             str
        fastq_2             str
        expected_cells      int"""
    )
    artifact.delete(permanent=True)


def test_record_soft_deleted_recreate():
    """Test that a soft-deleted record can be recreated with changes."""
    # testing soft delete and recreate with sqlite (postgres is tested in core/test_delete.py)
    # soft delete a record, then recreate it with some changes
    record = ln.Record(name="test_record").save()
    uid = record.uid
    assert record.branch_id == 1
    record.delete()
    assert record.branch_id == -1
    # now recreate the same record with the same uid but a different name
    record = ln.Record(name="test_record 2")
    record.uid = uid
    record.save()
    # now this record is recovered from the trash
    assert record.branch_id == 1
    assert record.name == "test_record 2"
    ln.Record.objects.filter().delete()



"""CELLxGENE utilities.

.. autosummary::
   :toctree: .

   save_cellxgene_defaults
   create_cellxgene_schema

"""

from ._cellxgene import (
    create_cellxgene_schema,
    get_cxg_schema,
    save_cellxgene_defaults,
    save_cxg_defaults,
)



# Generated by Django 5.2 on 2025-01-20 04:42

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.ids
import lamindb.base.users
import lamindb.models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0079_alter_rundata_value_json_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="persons",
        ),
        migrations.AddField(
            model_name="project",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="predecessors",
            field=models.ManyToManyField(
                related_name="successors", to="lamindb.project"
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=20,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_20,
                editable=False,
                max_length=20,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_8,
                editable=False,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.project",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_20,
                editable=False,
                max_length=20,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=20,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="space",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_default="00000000",
                db_index=True,
                default="00000000",
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="tidytable",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=16,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.ids.base62_8,
                editable=False,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=8,
                unique=True,
            ),
        ),
        migrations.CreateModel(
            name="PersonProject",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "role",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "person",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_project",
                        to="lamindb.person",
                    ),
                ),
                (
                    "project",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_person",
                        to="lamindb.project",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("person", "project")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="project",
            name="people",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.PersonProject",
                to="lamindb.person",
            ),
        ),
        migrations.AlterField(
            model_name="rundata",
            name="run",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="_rundata",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="paramvalue",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="started_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="runparamvalue",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="name",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="space",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="tidytable",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.CreateModel(
            name="RunULabel",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_ulabel",
                        to="lamindb.run",
                    ),
                ),
                (
                    "ulabel",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_run",
                        to="lamindb.ulabel",
                    ),
                ),
            ],
            options={
                "unique_together": {("run", "ulabel")},
            },
            bases=(models.Model, lamindb.models.IsLink),
        ),
        migrations.AddField(
            model_name="run",
            name="ulabels",
            field=models.ManyToManyField(
                related_name="runs", through="lamindb.RunULabel", to="lamindb.ulabel"
            ),
        ),
        migrations.RenameModel(
            old_name="TidyTable",
            new_name="FlexTable",
        ),
        migrations.RenameModel(
            old_name="TidyTableData",
            new_name="FlexTableData",
        ),
        migrations.RenameIndex(
            model_name="flextable",
            new_name="lamindb_fle_uid_e6f216_idx",
            old_name="lamindb_tid_uid_3a6e54_idx",
        ),
        migrations.RenameIndex(
            model_name="flextable",
            new_name="lamindb_fle_name_568594_idx",
            old_name="lamindb_tid_name_50c5de_idx",
        ),
        migrations.RenameIndex(
            model_name="flextabledata",
            new_name="lamindb_fle_tidytab_1674c1_idx",
            old_name="lamindb_tid_tidytab_b35a4d_idx",
        ),
        migrations.RenameIndex(
            model_name="flextabledata",
            new_name="lamindb_fle_feature_830e49_idx",
            old_name="lamindb_tid_feature_5a0b1f_idx",
        ),
        migrations.RenameIndex(
            model_name="flextabledata",
            new_name="lamindb_fle_param_i_4149cb_idx",
            old_name="lamindb_tid_param_i_16c884_idx",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_upath",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_upath",
        ),
    ]



import shutil
import sys
from pathlib import Path

docs_path = Path.cwd() / "docs" / "scripts"
sys.path.append(str(docs_path))

import bionty as bt
import lamindb as ln
import pandas as pd
import pytest
import tiledbsoma
import tiledbsoma.io
from lamindb.core import datasets
from lamindb.errors import InvalidArgument, ValidationError


@pytest.fixture(scope="module")
def mini_immuno_schema():
    # define labels
    perturbation = ln.ULabel(name="Perturbation", is_type=True).save()
    ln.ULabel(name="DMSO", type=perturbation).save()
    ln.ULabel(name="IFNG", type=perturbation).save()
    ln.ULabel(name="ulabel_but_not_perturbation").save()
    ln.ULabel.from_values(["sample1", "sample2", "sample3"], create=True).save()
    bt.CellType.from_source(name="B cell").save()
    bt.CellType.from_source(name="T cell").save()

    # in next iteration for attrs
    ln.Feature(name="temperature", dtype=float).save()
    # ln.Feature(name="experiment", dtype="cat[ULabel]").save()
    # ln.Feature(name="date_of_study", dtype="date").save()
    # ln.Feature(name="study_note", dtype="str").save()

    # define schema
    schema = ln.Schema(
        name="mini_immuno_obs_level_metadata_curator_tests",
        features=[
            ln.Feature(name="perturbation", dtype=perturbation).save(),
            ln.Feature(name="sample_note", dtype=str).save(),
            ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save(),
            ln.Feature(name="cell_type_by_model", dtype=bt.CellType).save(),
        ],
        index=ln.Feature(name="sample_label", dtype=ln.ULabel).save(),
    ).save()

    yield schema

    for af in ln.Artifact.filter():
        af.delete(permanent=True)

    from lamindb.models import SchemaComponent

    SchemaComponent.filter().delete(permanent=True)
    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)
    bt.Gene.filter().delete(permanent=True)
    ln.ULabel.filter(type__isnull=False).delete(permanent=True)
    ln.ULabel.filter().delete(permanent=True)
    bt.CellType.filter().delete(permanent=True)


@pytest.fixture(scope="module")
def curator_params():
    """Common curator parameters."""
    return {
        "categoricals": {
            "perturbation": ln.ULabel.name,
            "cell_type_by_expert": bt.CellType.name,
            "cell_type_by_model": bt.CellType.name,
        },
        "organism": "human",
    }


@pytest.fixture(scope="module")
def mudata_papalexi21_subset_schema():
    # define labels
    perturbation = ln.ULabel(name="Perturbation", is_type=True).save()
    ln.ULabel(name="Perturbed", type=perturbation).save()
    ln.ULabel(name="NT", type=perturbation).save()

    replicate = ln.ULabel(name="Replicate", is_type=True).save()
    ln.ULabel(name="rep1", type=replicate).save()
    ln.ULabel(name="rep2", type=replicate).save()
    ln.ULabel(name="rep3", type=replicate).save()

    # define obs schema
    obs_schema = ln.Schema(
        name="mudata_papalexi21_subset_obs_schema",
        features=[
            ln.Feature(name="perturbation", dtype="cat[ULabel[Perturbation]]").save(),
            ln.Feature(name="replicate", dtype="cat[ULabel[Replicate]]").save(),
        ],
    ).save()

    obs_schema_rna = ln.Schema(
        name="mudata_papalexi21_subset_rna_obs_schema",
        features=[
            ln.Feature(name="nCount_RNA", dtype=int).save(),
            ln.Feature(name="nFeature_RNA", dtype=int).save(),
            ln.Feature(name="percent.mito", dtype=float).save(),
        ],
        coerce_dtype=True,
    ).save()

    obs_schema_hto = ln.Schema(
        name="mudata_papalexi21_subset_hto_obs_schema",
        features=[
            ln.Feature(name="nCount_HTO", dtype=float).save(),
            ln.Feature(name="nFeature_HTO", dtype=int).save(),
            ln.Feature(name="technique", dtype=bt.ExperimentalFactor).save(),
        ],
        coerce_dtype=True,
    ).save()

    var_schema_rna = ln.Schema(
        name="mudata_papalexi21_subset_rna_var_schema",
        itype=bt.Gene.symbol,
        dtype=float,
    ).save()

    # define composite schema
    mudata_schema = ln.Schema(
        name="mudata_papalexi21_subset_mudata_schema",
        otype="MuData",
        slots={
            "obs": obs_schema,
            "rna:obs": obs_schema_rna,
            "hto:obs": obs_schema_hto,
            "rna:var": var_schema_rna,
        },
    ).save()

    yield mudata_schema

    for af in ln.Artifact.filter():
        af.delete(permanent=True)
    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)
    bt.models.SchemaGene.filter().delete()
    bt.Gene.filter().delete(permanent=True)
    ln.ULabel.filter(type__isnull=False).delete(permanent=True)
    ln.ULabel.filter().delete(permanent=True)
    bt.ExperimentalFactor.filter().delete(permanent=True)


@pytest.fixture(scope="module")
def study_metadata_schema():
    from define_schema_df_metadata import study_metadata_schema

    yield study_metadata_schema

    study_metadata_schema.delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)


@pytest.fixture(scope="module")
def anndata_uns_schema():
    from define_schema_anndata_uns import anndata_uns_schema

    yield anndata_uns_schema

    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)


@pytest.fixture(scope="module")
def spatialdata_blobs_schema():
    from define_schema_spatialdata import sdata_schema

    yield sdata_schema

    for af in ln.Artifact.filter():
        af.delete(permanent=True)

    from lamindb.models import SchemaComponent

    SchemaComponent.filter().delete(permanent=True)

    ln.Schema.filter().delete(permanent=True)
    bt.models.SchemaGene.filter().delete()
    bt.Gene.filter().delete(permanent=True)
    ln.ULabel.filter(type__isnull=False).delete(permanent=True)
    ln.ULabel.filter().delete(permanent=True)
    bt.ExperimentalFactor.filter().delete(permanent=True)
    bt.DevelopmentalStage.filter().delete(permanent=True)
    bt.Disease.filter().delete(permanent=True)


def test_dataframe_curator(mini_immuno_schema: ln.Schema):
    """Test DataFrame curator implementation."""

    ln.settings.verbosity = "info"

    # invalid simple dtype (float)
    feature_to_fail = ln.Feature(name="treatment_time_h", dtype=float).save()
    schema = ln.Schema(
        name="mini_immuno_obs_level_metadata_v2",
        features=[
            ln.Feature(name="perturbation", dtype="cat[ULabel[Perturbation]]").save(),
            ln.Feature(name="sample_note", dtype=str).save(),
            ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save(),
            ln.Feature(name="cell_type_by_model", dtype=bt.CellType).save(),
            feature_to_fail,
        ],
    ).save()
    df = datasets.mini_immuno.get_dataset1(otype="DataFrame")
    curator = ln.curators.DataFrameCurator(df, schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert (
        "Column 'treatment_time_h' failed series or dataframe validator 0: <Check check_function: Column 'treatment_time_h' failed dtype check for 'float': got int64>"
        in error.exconly()
    )

    schema.delete(permanent=True)
    feature_to_fail.delete(permanent=True)

    # Wrong subtype
    df = datasets.mini_immuno.get_dataset1(otype="DataFrame", with_wrong_subtype=True)
    curator = ln.curators.DataFrameCurator(df, mini_immuno_schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: 1 term not validated in feature 'perturbation': 'ulabel_but_not_perturbation'
    → fix typos, remove non-existent values, or save terms via: curator.cat.add_new_from('perturbation')
    → a valid label for subtype 'Perturbation' has to be one of ['DMSO', 'IFNG']"""
    )

    # Typo
    df = datasets.mini_immuno.get_dataset1(otype="DataFrame", with_typo=True)
    curator = ln.curators.DataFrameCurator(df, mini_immuno_schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: 1 term not validated in feature 'perturbation': 'IFNJ'
    → fix typos, remove non-existent values, or save terms via: curator.cat.add_new_from('perturbation')
    → a valid label for subtype 'Perturbation' has to be one of ['DMSO', 'IFNG']"""
    )

    df = datasets.mini_immuno.get_dataset1(otype="DataFrame")
    curator = ln.curators.DataFrameCurator(df, mini_immuno_schema)
    artifact = curator.save_artifact(key="examples/dataset1.parquet")

    assert artifact.schema == mini_immuno_schema
    assert artifact.features.slots["columns"].n == 5
    assert (
        artifact.features.describe(return_str=True)
        == """\
Artifact .parquet · DataFrame · dataset
└── Dataset features
    └── columns • 5         [Feature]
        cell_type_by_expe…  cat[bionty.CellType]    B cell, CD8-positive, alpha…
        cell_type_by_model  cat[bionty.CellType]    B cell, T cell
        perturbation        cat[ULabel[Perturbati…  DMSO, IFNG
        sample_label        cat[ULabel]             sample1, sample2, sample3
        sample_note         str"""
    )
    assert set(artifact.features.get_values()["sample_label"]) == {
        "sample1",
        "sample2",
        "sample3",
    }
    assert set(artifact.features.get_values()["cell_type_by_expert"]) == {
        "CD8-positive, alpha-beta T cell",
        "B cell",
    }
    assert set(artifact.features.get_values()["cell_type_by_model"]) == {
        "T cell",
        "B cell",
    }

    # a second dataset with missing values
    ln.ULabel.from_values(["sample4", "sample5", "sample6"], create=True).save()
    df = ln.examples.datasets.mini_immuno.get_dataset2(
        otype="DataFrame", gene_symbols_in_index=True
    )
    curator = ln.curators.DataFrameCurator(df, mini_immuno_schema)
    try:
        curator.validate()
    except ln.errors.ValidationError as error:
        assert "column 'sample_note' not in dataframe" in str(error)
    curator.standardize()
    curator.validate()

    artifact.delete(permanent=True)


def test_dataframe_curator_index():
    """Test validating a DataFrame index."""
    df = datasets.mini_immuno.get_dataset1(
        otype="DataFrame", with_index_type_mismatch=True
    )
    feature = ln.Feature(name="test", dtype="str").save()
    schema = ln.Schema(index=feature).save()
    curator = ln.curators.DataFrameCurator(df, schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert "expected series 'None' to have type str" in error.exconly()

    schema.delete(permanent=True)
    feature.delete(permanent=True)


def test_dataframe_curator_validate_all_annotate_cat(mini_immuno_schema):
    """Do not pass any features."""
    schema = ln.Schema(itype=ln.Feature).save()
    assert schema.flexible

    df = datasets.mini_immuno.get_dataset1(otype="DataFrame")
    artifact = ln.Artifact.from_dataframe(
        df, key="examples/dataset1.parquet", schema=schema
    ).save()
    assert set(artifact.features.get_values()["perturbation"]) == {
        "DMSO",
        "IFNG",
    }
    assert set(artifact.features.get_values()["cell_type_by_expert"]) == {
        "CD8-positive, alpha-beta T cell",
        "B cell",
    }
    assert set(artifact.features.get_values()["cell_type_by_model"]) == {
        "T cell",
        "B cell",
    }

    artifact.delete(permanent=True)
    schema.delete(permanent=True)


def test_dataframe_curator_validate_all_annotate_cat2(mini_immuno_schema):
    """Combine half-specifying features, half not."""
    schema = ln.Schema(
        itype=ln.Feature,
        features=[ln.Feature.get(name="perturbation")],
        flexible=True,
    ).save()
    assert schema.flexible

    df = datasets.mini_immuno.get_dataset1(otype="DataFrame")
    curator = ln.curators.DataFrameCurator(df, schema)
    artifact = curator.save_artifact(key="examples/dataset1.parquet")
    assert set(artifact.features.get_values()["perturbation"]) == {
        "DMSO",
        "IFNG",
    }
    assert set(artifact.features.get_values()["cell_type_by_expert"]) == {
        "CD8-positive, alpha-beta T cell",
        "B cell",
    }
    assert set(artifact.features.get_values()["cell_type_by_model"]) == {
        "T cell",
        "B cell",
    }

    artifact.delete(permanent=True)
    schema.delete(permanent=True)


@pytest.mark.parametrize("include_attrs_slot", [True, False])
def test_dataframe_attrs_validation(study_metadata_schema, include_attrs_slot):
    df = datasets.mini_immuno.get_dataset1(otype="DataFrame")

    perturbation = ln.ULabel(name="Perturbation", is_type=True).save()
    perturbation_feature = ln.Feature(name="perturbation", dtype=perturbation).save()
    ln.ULabel(name="DMSO", type=perturbation).save()
    ln.ULabel(name="IFNG", type=perturbation).save()

    if include_attrs_slot:
        schema = ln.Schema(
            features=[perturbation_feature],
            slots={"attrs": study_metadata_schema},
            otype="DataFrame",
        ).save()
    else:
        schema = ln.Schema(
            features=[perturbation_feature],
            otype="DataFrame",
        ).save()

    bad_schema = ln.Schema(slots={"doesnotexist": schema}, otype="DataFrame").save()

    with pytest.raises(ValueError) as e:
        curator = ln.curators.DataFrameCurator(df, schema=bad_schema)
    assert (
        "Slot 'doesnotexist' is not supported for DataFrameCurator. Must be 'attrs'."
        in str(e.value)
    )

    curator = ln.curators.DataFrameCurator(df, schema=schema)

    if include_attrs_slot:
        assert curator.slots["attrs"].__class__.__name__ == "ComponentCurator"
    else:
        assert not curator.slots

    curator.validate()
    artifact = curator.save_artifact(key="examples/df_with_attrs.parquet")

    assert artifact.schema == schema
    if include_attrs_slot:
        assert "attrs" in artifact.features.slots
        assert artifact.features.slots["attrs"].features.first() == ln.Feature.get(
            name="temperature"
        )
        assert artifact.features.slots["attrs"].features.last() == ln.Feature.get(
            name="experiment"
        )
    else:
        assert (
            not hasattr(artifact.features, "slots")
            or "attrs" not in artifact.features.slots
        )

    from lamindb.models import SchemaComponent

    SchemaComponent.filter().delete(permanent=True)
    artifact.delete(permanent=True)
    bad_schema.delete(permanent=True)
    schema.delete(permanent=True)


def test_schema_new_genes(ccaplog):
    df = pd.DataFrame(
        index=pd.Index(
            [
                "ENSG00000139618",  # BRCA2
                "ENSG00000141510",  # TP53
                "ENSG00999000001",  # Invalid ID
                "ENSG00999000002",  # Invalid ID
            ],
            name="ensembl",
        )
    )
    feature = ln.Feature(name="ensembl", dtype=bt.Gene.ensembl_gene_id).save()
    schema = ln.Schema(index=feature).save()
    curator = ln.curators.DataFrameCurator(df, schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert error.exconly().startswith(
        "lamindb.errors.ValidationError: 2 terms not validated in feature 'index': 'ENSG00999000001', 'ENSG00999000002'"
    )

    assert (
        "2 terms not validated in feature 'index': 'ENSG00999000001', 'ENSG00999000002'"
        in ccaplog.text
    )

    schema.delete(permanent=True)
    feature.delete(permanent=True)


def test_schema_no_match_ensembl():
    df = pd.DataFrame(
        index=pd.Index(
            [
                "ENSG99999999998",  # Invalid ID
                "ENSG99999999999",  # Invalid ID
            ],
            name="ensembl",
        )
    )
    schema = ln.Schema(
        index=ln.Feature(name="ensembl", dtype=bt.Gene.ensembl_gene_id).save()
    ).save()
    curator = ln.curators.DataFrameCurator(df, schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: 2 terms not validated in feature 'index': 'ENSG99999999998', 'ENSG99999999999'
    → fix organism 'human', fix typos, remove non-existent values, or save terms via: curator.cat.add_new_from('index')"""
    )

    schema.delete(permanent=True)


def test_schema_mixed_ensembl_symbols(ccaplog):
    """Quite some datasets have mixed ensembl gene IDs and symbols.

    The expected behavior is that an error is raised when such a dataset is encountered because
    currently lamindb does not support validating values against a union of Fields.

    The current behavior is that these cases automatically pass.
    """
    df = pd.DataFrame(
        index=pd.Index(
            [
                "ENSG00000139618",
                "ENSG00000141510",
                "BRCA2",  # symbol
                "TP53",  # symbol
            ],
            name="ensembl",
        )
    )
    schema = ln.Schema(
        index=ln.Feature(name="ensembl", dtype=bt.Gene.ensembl_gene_id).save()
    ).save()
    curator = ln.curators.DataFrameCurator(df, schema)
    with pytest.raises(ln.errors.ValidationError) as error:
        curator.validate()
    assert error.exconly().startswith(
        "lamindb.errors.ValidationError: 2 terms not validated in feature 'index': 'BRCA2', 'TP53'"
    )

    assert "2 terms not validated in feature 'index': 'BRCA2', 'TP53'" in ccaplog.text

    schema.delete(permanent=True)


def test_anndata_curator_different_components(mini_immuno_schema: ln.Schema):
    obs_schema = mini_immuno_schema

    for add_comp in ["var.T", "obs", "uns"]:
        var_schema = ln.Schema(
            name="scRNA_seq_var_schema",
            itype=bt.Gene.ensembl_gene_id,
            dtype="num",
        ).save()

        # always assume var
        components = {"var.T": var_schema}
        if add_comp == "obs":
            components["obs"] = obs_schema
        if add_comp == "uns":
            uns_schema = ln.Schema(
                name="flexible_uns_schema",
                itype=ln.Feature,
            ).save()
            components["uns"] = uns_schema

        anndata_schema = ln.Schema(
            name="mini_immuno_anndata_schema",
            otype="AnnData",
            slots=components,
        ).save()
        assert mini_immuno_schema.id is not None, mini_immuno_schema
        assert anndata_schema.slots["var.T"] == var_schema
        if add_comp == "obs":
            assert anndata_schema.slots["obs"] == obs_schema
        if add_comp == "uns":
            assert anndata_schema.slots["uns"] == uns_schema

        describe_output = anndata_schema.describe(return_str=True)
        assert "mini_immuno_anndata_schema" in describe_output
        assert "scRNA_seq_var_schema" in describe_output
        if add_comp == "obs":
            assert "mini_immuno_anndata_schema" in describe_output
        if add_comp == "uns":
            assert "flexible_uns_schema" in describe_output

        adata = datasets.mini_immuno.get_dataset1(otype="AnnData")
        curator = ln.curators.AnnDataCurator(adata, anndata_schema)
        assert curator.slots["var.T"].__class__.__name__ == "ComponentCurator"
        if add_comp == "obs":
            assert curator.slots["obs"].__class__.__name__ == "ComponentCurator"
        if add_comp == "uns":
            assert curator.slots["uns"].__class__.__name__ == "ComponentCurator"

        artifact = ln.Artifact.from_anndata(
            adata, key="examples/dataset1.h5ad", schema=anndata_schema
        )
        assert artifact._curator._is_validated  # important test, do not remove
        artifact.save()
        assert not hasattr(artifact, "_curator")  # test that curator is deleted
        assert artifact.schema == anndata_schema
        assert artifact.features.slots["var.T"].n == 3  # 3 genes get linked
        if add_comp == "obs":
            assert artifact.features.slots["obs"] == obs_schema
            # deprecated
            assert artifact.features._schema_by_slot["obs"] == obs_schema
            assert artifact.features._feature_set_by_slot["obs"] == obs_schema

            assert set(artifact.features.get_values()["cell_type_by_expert"]) == {
                "CD8-positive, alpha-beta T cell",
                "B cell",
            }
            assert set(artifact.features.get_values()["cell_type_by_model"]) == {
                "T cell",
                "B cell",
            }
        if add_comp == "uns":
            assert artifact.features.slots["uns"].features.first() == ln.Feature.get(
                name="temperature"
            )

        artifact.delete(permanent=True)
        anndata_schema.delete(permanent=True)
        var_schema.delete(permanent=True)


def test_anndata_curator_varT_curation():
    ln.Schema.filter(itype="bionty.Gene.ensembl_gene_id").delete()
    varT_schema = ln.Schema(itype=bt.Gene.ensembl_gene_id, maximal_set=True).save()
    slot = "var.T"
    components = {slot: varT_schema}
    anndata_schema = ln.Schema(
        otype="AnnData",
        slots=components,
    ).save()
    for with_gene_typo in [True, False]:
        adata = datasets.mini_immuno.get_dataset1(
            otype="AnnData", with_gene_typo=with_gene_typo
        )
        if with_gene_typo:
            with pytest.raises(ValidationError) as error:
                artifact = ln.Artifact.from_anndata(
                    adata, key="examples/dataset1.h5ad", schema=anndata_schema
                ).save()
            assert error.exconly() == (
                f"lamindb.errors.ValidationError: 1 term not validated in feature 'columns' in slot '{slot}': 'GeneTypo'\n"
                f"    → fix organism 'human', fix typos, remove non-existent values, or save terms via: curator.slots['{slot}'].cat.add_new_from('columns')"
            )
        else:
            for n_max_records in [2, 4]:
                ln.settings.annotation.n_max_records = n_max_records
                artifact = ln.Artifact.from_anndata(
                    adata, key="examples/dataset1.h5ad", schema=anndata_schema
                ).save()
                assert artifact.features.slots[slot].n == 3  # 3 genes get linked
                assert (
                    artifact.features.slots[slot].itype == "bionty.Gene.ensembl_gene_id"
                )
                if n_max_records == 2:
                    assert not artifact.features.slots[slot].members.exists()
                else:
                    assert artifact.features.slots[slot].members.to_dataframe()[
                        "ensembl_gene_id"
                    ].tolist() == [
                        "ENSG00000153563",
                        "ENSG00000010610",
                        "ENSG00000170458",
                    ]

                artifact.delete(permanent=True)

            anndata_schema.delete(permanent=True)
            varT_schema.delete(permanent=True)


def test_anndata_curator_varT_curation_legacy(ccaplog):
    varT_schema = ln.Schema(itype=bt.Gene.ensembl_gene_id, maximal_set=True).save()
    slot = "var"
    components = {slot: varT_schema}
    anndata_schema = ln.Schema(
        otype="AnnData",
        slots=components,
    ).save()
    for with_gene_typo in [True, False]:
        adata = datasets.mini_immuno.get_dataset1(
            otype="AnnData", with_gene_typo=with_gene_typo
        )
        if with_gene_typo:
            with pytest.raises(ValidationError) as error:
                artifact = ln.Artifact.from_anndata(
                    adata, key="examples/dataset1.h5ad", schema=anndata_schema
                ).save()
            assert error.exconly() == (
                f"lamindb.errors.ValidationError: 1 term not validated in feature 'var_index' in slot '{slot}': 'GeneTypo'\n"
                f"    → fix organism 'human', fix typos, remove non-existent values, or save terms via: curator.slots['{slot}'].cat.add_new_from('var_index')"
            )
        else:
            artifact = ln.Artifact.from_anndata(
                adata, key="examples/dataset1.h5ad", schema=anndata_schema
            ).save()
            assert (
                "auto-transposed `var` for backward compat, please indicate transposition in the schema definition by calling out `.T`: slots={'var.T': itype=bt.Gene.ensembl_gene_id}"
                in ccaplog.text
            )
            assert artifact.features.slots[slot].n == 3  # 3 genes get linked
            assert set(
                artifact.features.slots[slot].members.to_dataframe()["ensembl_gene_id"]
            ) == {
                "ENSG00000153563",
                "ENSG00000010610",
                "ENSG00000170458",
            }

            artifact.delete(permanent=True)

            anndata_schema.delete(permanent=True)
            varT_schema.delete(permanent=True)


def test_anndata_curator_nested_uns(study_metadata_schema, anndata_uns_schema):
    """Test AnnDataCurator with nested uns slot validation."""
    adata = datasets.mini_immuno.get_dataset1(otype="AnnData")
    adata.uns["study_metadata"] = adata.uns.copy()

    curator = ln.curators.AnnDataCurator(adata, anndata_uns_schema)
    assert curator.slots["uns:study_metadata"].__class__.__name__ == "ComponentCurator"

    curator.validate()
    artifact = curator.save_artifact(key="examples/anndata_with_uns.h5ad")

    assert artifact.schema == anndata_uns_schema
    assert "uns:study_metadata" in artifact.features.slots
    assert artifact.features.slots[
        "uns:study_metadata"
    ].features.first() == ln.Feature.get(name="temperature")

    adata = datasets.mini_immuno.get_dataset1(otype="AnnData")
    bad_schema1 = ln.Schema(
        otype="AnnData",
        slots={"uns:nonexistent": study_metadata_schema},
    ).save()
    with pytest.raises(InvalidArgument) as e:
        ln.curators.AnnDataCurator(adata, bad_schema1)
    assert (
        "Schema slot 'uns:nonexistent' requires keys uns['nonexistent'] but key 'nonexistent' not found."
        in str(e.value)
    )

    with pytest.raises(InvalidArgument) as e:
        bad_schema2 = ln.Schema(
            otype="AnnData",
            slots={"uns:temperature:nonexistent_nested": study_metadata_schema},
        ).save()
        ln.curators.AnnDataCurator(adata, bad_schema2)
    assert (
        "Schema slot 'uns:temperature:nonexistent_nested' requires keys uns['temperature']['nonexistent_nested'] but key 'nonexistent_nested' not found. Available keys at this level: none (not a dict)."
        in str(e.value)
    )

    inferred_sets = artifact.feature_sets.all()
    for inferred_set in inferred_sets:
        artifact.feature_sets.remove(inferred_set)
    artifact.delete(permanent=True)
    bad_schema1.delete(permanent=True)
    bad_schema2.delete(permanent=True)
    anndata_uns_schema.delete(permanent=True)


def test_anndata_curator_no_var(mini_immuno_schema: ln.Schema):
    assert mini_immuno_schema.id is not None, mini_immuno_schema
    # test no var schema
    anndata_schema_no_var = ln.Schema(
        name="mini_immuno_anndata_schema_no_var",
        otype="AnnData",
        slots={"obs": mini_immuno_schema},
    ).save()
    assert mini_immuno_schema.id is not None, mini_immuno_schema
    adata = datasets.mini_immuno.get_dataset1(otype="AnnData")
    curator = ln.curators.AnnDataCurator(adata, anndata_schema_no_var)

    artifact = curator.save_artifact(key="examples/dataset1_no_var.h5ad")
    artifact.delete(permanent=True)
    anndata_schema_no_var.delete(permanent=True)


def test_mudata_curator(
    mudata_papalexi21_subset_schema: ln.Schema, mini_immuno_schema: ln.Schema
):
    mudata_schema = mudata_papalexi21_subset_schema
    mdata = ln.examples.datasets.mudata_papalexi21_subset()
    # TODO: refactor organism
    bt.settings.organism = "human"
    # wrong dataset
    with pytest.raises(InvalidArgument):
        ln.curators.MuDataCurator(pd.DataFrame(), mudata_schema)
    # wrong schema
    with pytest.raises(InvalidArgument):
        ln.curators.MuDataCurator(mdata, mini_immuno_schema)
    curator = ln.curators.MuDataCurator(mdata, mudata_schema)
    assert curator.slots.keys() == {
        "obs",
        "rna:obs",
        "hto:obs",
        "rna:var",
    }
    ln.settings.verbosity = "hint"
    curator.validate()
    curator.slots["rna:var"].cat.standardize("columns")
    curator.slots["rna:var"].cat.add_new_from("columns")
    artifact = curator.save_artifact(key="mudata_papalexi21_subset.h5mu")
    assert artifact.schema == mudata_schema
    assert set(artifact.features.slots.keys()) == {
        "obs",
        "rna:var",
        "rna:obs",
        "hto:obs",
    }

    artifact.delete(permanent=True)
    mudata_schema.delete(permanent=True)
    mini_immuno_schema.delete(permanent=True)
    Path("papalexi21_subset.h5mu").unlink(missing_ok=True)


def test_mudata_curator_nested_uns(study_metadata_schema):
    """Test MuData with nested uns slot validation.

    This test verifies the behavior of both the MuData `.uns` slots and a `.uns` slot of
    an AnnData object inside the MuData object that gets specified using the key `:` syntax.
    """
    mdata = ln.examples.datasets.mudata_papalexi21_subset(with_uns=True)

    site_uns_schema = ln.Schema(
        features=[
            ln.Feature(name="pos", dtype=float).save(),
            ln.Feature(name="site_id", dtype=str).save(),
        ]
    ).save()

    mdata_schema = ln.Schema(
        otype="MuData",
        slots={
            "uns:study_metadata": study_metadata_schema,
            "rna:uns:site_metadata": site_uns_schema,
        },
    ).save()

    curator = ln.curators.MuDataCurator(mdata, mdata_schema)
    assert curator.slots["uns:study_metadata"].__class__.__name__ == "ComponentCurator"
    assert (
        curator.slots["rna:uns:site_metadata"].__class__.__name__ == "ComponentCurator"
    )

    curator.validate()
    artifact = curator.save_artifact(key="examples/mdata_with_uns.h5mu")

    assert artifact.schema == mdata_schema
    assert "uns:study_metadata" in artifact.features.slots
    assert "rna:uns:site_metadata" in artifact.features.slots
    assert artifact.features.slots[
        "uns:study_metadata"
    ].features.first() == ln.Feature.get(name="temperature")
    assert artifact.features.slots[
        "rna:uns:site_metadata"
    ].features.first() == ln.Feature.get(name="pos")

    # Clean up
    artifact.delete(permanent=True)
    Path("papalexi21_subset.h5mu").unlink(missing_ok=True)


def test_spatialdata_curator(
    spatialdata_blobs_schema: ln.Schema,
):
    spatialdata = ln.examples.datasets.spatialdata_blobs()

    # wrong dataset
    with pytest.raises(InvalidArgument):
        ln.curators.SpatialDataCurator(pd.DataFrame(), spatialdata_blobs_schema)
    # wrong schema - use an actual slot that exists
    with pytest.raises(InvalidArgument):
        ln.curators.SpatialDataCurator(
            spatialdata, spatialdata_blobs_schema.slots["attrs:bio"]
        )

    curator = ln.curators.SpatialDataCurator(spatialdata, spatialdata_blobs_schema)
    with pytest.raises(ln.errors.ValidationError):
        curator.validate()

    spatialdata.tables["table"].var.drop(index="ENSG00000999999", inplace=True)
    artifact = ln.Artifact.from_spatialdata(
        spatialdata,
        key="examples/spatialdata1.zarr",
        schema=spatialdata_blobs_schema,
    ).save()
    assert artifact.schema == spatialdata_blobs_schema
    assert artifact.features.slots.keys() == {
        "attrs:bio",
        "attrs:tech",
        "attrs",
        "tables:table:obs",
        "tables:table:var.T",
    }
    assert artifact.features.get_values()["disease"] == "Alzheimer disease"

    assert (
        artifact.features.describe(return_str=True)
        == """Artifact .zarr · SpatialData · dataset
└── Dataset features
    ├── attrs:bio • 2       [Feature]
    │   developmental_sta…  cat[bionty.Developmen…  adult stage
    │   disease             cat[bionty.Disease]     Alzheimer disease
    ├── attrs:tech • 1      [Feature]
    │   assay               cat[bionty.Experiment…  Visium Spatial Gene Express…
    ├── attrs • 2           [Feature]
    │   bio                 dict
    │   tech                dict
    ├── tables:table:obs …  [Feature]
    │   sample_region       str
    └── tables:table:var.…  [bionty.Gene.ensembl_…
        BRCA2               num
        BRAF                num"""
    )

    artifact.delete(permanent=True)


def test_tiledbsoma_curator(clean_soma_files):
    """Test TiledbSomaExperimentCurator with schema."""
    obs_schema = ln.Schema(
        features=[
            ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save(),
            ln.Feature(name="cell_type_by_model", dtype=bt.CellType).save(),
        ],
    ).save()

    var_schema = ln.Schema(
        features=[
            ln.Feature(name="var_id", dtype=bt.Gene.ensembl_gene_id).save(),
        ],
        coerce_dtype=True,
    ).save()

    soma_schema = ln.Schema(
        otype="tiledbsoma",
        slots={
            "obs": obs_schema,
            "ms:RNA": var_schema,
        },
    ).save()

    # Convert AnnData to SOMA format
    adata = ln.examples.datasets.mini_immuno.get_dataset1(otype="AnnData")
    tiledbsoma.io.from_anndata(
        "small_dataset.tiledbsoma", adata, measurement_name="RNA"
    )

    # Test with invalid dataset
    with pytest.raises(ln.errors.InvalidArgument) as e:
        ln.curators.TiledbsomaExperimentCurator(adata, soma_schema)
    assert "dataset must be SOMAExperiment-like." in str(e.value)

    # Test with invalid schema
    with tiledbsoma.Experiment.open("small_dataset.tiledbsoma") as experiment:
        with pytest.raises(ln.errors.InvalidArgument) as e:
            ln.curators.TiledbsomaExperimentCurator(experiment, schema=var_schema)
        assert "Schema otype must be 'tiledbsoma'." in str(e.value)

    with tiledbsoma.Experiment.open("small_dataset.tiledbsoma") as experiment:
        curator = ln.curators.TiledbsomaExperimentCurator(experiment, soma_schema)

        assert "obs" in curator.slots
        assert "ms:RNA" in curator.slots

        curator.validate()

        artifact = curator.save_artifact(
            key="examples/soma_experiment.tiledbsoma",
            description="SOMA experiment with schema validation",
        )

        assert artifact.schema == soma_schema
        assert "obs" in artifact.features.slots
        assert "ms:RNA" in artifact.features.slots

        # Check feature values are properly annotated
        assert set(artifact.features.get_values()["cell_type_by_expert"]) == {
            "CD8-positive, alpha-beta T cell",
            "B cell",
        }
        assert set(artifact.features.get_values()["cell_type_by_model"]) == {
            "T cell",
            "B cell",
        }

    # Altered data (gene typo)
    adata_typo = ln.examples.datasets.mini_immuno.get_dataset1(
        otype="AnnData", with_gene_typo=True
    )
    typo_soma_path = "./mini_immuno_dataset1_typo.tiledbsoma"
    tiledbsoma.io.from_anndata(typo_soma_path, adata_typo, measurement_name="RNA")
    with tiledbsoma.Experiment.open(typo_soma_path) as experiment_typo:
        curator_typo = ln.curators.TiledbsomaExperimentCurator(
            experiment_typo, soma_schema
        )

        # Validation should fail due to typo
        with pytest.raises(ln.errors.ValidationError) as error:
            curator_typo.validate()
        assert "GeneTypo" in str(error.value)

    # Clean up
    shutil.rmtree(typo_soma_path)
    artifact.delete(permanent=True)
    soma_schema.delete(permanent=True)
    var_schema.delete(permanent=True)
    obs_schema.delete(permanent=True)



import shutil

import anndata as ad
import lamindb as ln
import pytest
from lamindb.errors import (
    IntegrityError,
)


def test_create_from_anndata_in_existing_cloud_storage():
    filepath = "s3://lamindb-test/core/scrnaseq_pbmc68k_tiny.h5ad"
    artifact = ln.Artifact.from_anndata(
        filepath, description="test_create_from_anndata_cloudpath"
    )
    assert artifact.n_observations == 70
    artifact.save()
    assert ln.Artifact.get(path=artifact.path) == artifact
    # check that the local filepath has been cleared
    assert not hasattr(artifact, "_local_filepath")
    assert artifact.path.as_posix().startswith("s3://lamindb-test/core")


@pytest.mark.parametrize(
    "filepath_str",
    ["s3://lamindb-ci/test-data/test.parquet", "s3://lamindb-ci/test-data/test.csv"],
)
@pytest.mark.parametrize("skip_check_exists", [False, True])
@pytest.mark.parametrize("skip_size_and_hash", [False, True])
def test_create_small_file_from_remote_path(
    filepath_str, skip_check_exists, skip_size_and_hash
):
    ln.settings.creation.artifact_skip_size_hash = skip_size_and_hash
    artifact = ln.Artifact(
        filepath_str,
        skip_check_exists=skip_check_exists,
    )
    artifact.save()
    # test cache()
    file_from_local = ln.Artifact(artifact.cache(), description="test")
    # test hash equivalency when computed on local machine
    if not skip_size_and_hash:
        assert file_from_local.hash == artifact.hash
        assert file_from_local._hash_type == "md5"
        assert artifact._hash_type == "md5"
    assert artifact.path.as_posix() == filepath_str
    assert artifact.load().iloc[0].tolist() == [
        0,
        "Abingdon island giant tortoise",
        "Chelonoidis abingdonii",
        106734,
        "ASM359739v1",
        "GCA_003597395.1",
        "Full genebuild",
        "-",
        "-",
    ]
    artifact.delete(permanent=True, storage=False)
    ln.settings.creation.artifact_skip_size_hash = False


def test_versioning_arifact_from_existing_path():
    artifact1 = ln.Artifact("s3://lamindb-ci/test-data/test.parquet").save()
    artifact2 = ln.Artifact(
        "s3://lamindb-ci/test-data/test.csv", revises=artifact1
    ).save()
    assert artifact1.stem_uid == artifact2.stem_uid
    assert artifact1.uid != artifact2.uid
    artifact1.delete(permanent=True, storage=False)
    artifact2.delete(permanent=True, storage=False)


def test_create_big_file_from_remote_path():
    # the point of this test is check the multi-upload hash
    filepath_str = "s3://lamindb-test/core/human_immune.h5ad"
    # we don't use from_anndata() here because we test this with a small file for shorter run time
    artifact = ln.Artifact(filepath_str)
    assert not artifact._key_is_virtual
    assert artifact._real_key is None
    assert artifact.key == "human_immune.h5ad"
    assert artifact._hash_type == "md5-3"
    assert artifact.size == 21960324
    assert artifact.path.as_posix() == filepath_str
    # check _real_key
    artifact = ln.Artifact(filepath_str, key="adata_test_key.h5ad")
    assert artifact._key_is_virtual
    assert artifact.key == "adata_test_key.h5ad"
    assert artifact._real_key.endswith("human_immune.h5ad")
    assert artifact.path.as_posix() == filepath_str


def test_delete_artifact_from_non_managed_storage():
    artifact = ln.Artifact(
        "s3://lamindb-dev-datasets/file-to-test-for-delete.csv",
        description="My test file to delete from non-default storage",
    ).save()
    assert artifact.storage.instance_uid != ln.setup.settings.instance.uid
    assert artifact.key is not None
    filepath = artifact.path
    with pytest.raises(IntegrityError) as e:
        artifact.delete()
    assert e.exconly().startswith(
        "lamindb.errors.IntegrityError: Cannot simply delete artifacts"
    )
    artifact.delete(storage=False, permanent=True)
    assert (
        ln.Artifact.filter(
            description="My test file to delete from non-default storage",
            branch_id=None,
        ).first()
        is None
    )
    assert filepath.exists()


def test_huggingface_paths():
    artifact_adata = ln.Artifact(
        "hf://datasets/Koncopd/lamindb-test@main/anndata/pbmc68k_test.h5ad",
        description="hf adata",
    )
    artifact_adata.save()
    assert artifact_adata.hash is not None
    assert isinstance(artifact_adata.load(), ad.AnnData)
    assert artifact_adata._cache_path.exists()
    artifact_adata._cache_path.unlink()

    artifact_pq = ln.Artifact(
        "hf://datasets/Koncopd/lamindb-test/sharded_parquet", description="hf parquet"
    )
    artifact_pq.save()
    assert artifact_pq.hash is not None
    assert len(artifact_pq.open().files) == 11
    assert artifact_pq.cache().is_dir()
    shutil.rmtree(artifact_pq._cache_path)

    artifact_adata.delete(permanent=True, storage=False)
    artifact_pq.delete(permanent=True, storage=False)


def test_gcp_paths():
    artifact_folder = ln.Artifact(
        "gs://rxrx1-europe-west4/images/test/HEPG2-08", description="Test GCP folder"
    ).save()
    assert artifact_folder.hash == "6r5Hkce0UTy7X6gLeaqzBA"
    assert artifact_folder.n_files == 14772

    artifact_file = ln.Artifact(
        "gs://rxrx1-europe-west4/images/test/HEPG2-08/Plate1/B02_s1_w1.png",
        description="Test GCP file",
    ).save()
    assert artifact_file.hash == "foEgLjmuUHO62CazxN97rA"
    cache_path = artifact_file.cache()
    assert cache_path.is_file()

    cache_path.unlink()
    artifact_folder.delete(permanent=True, storage=False)
    artifact_file.delete(permanent=True, storage=False)


def test_http_paths():
    http_path = ln.UPath(
        "https://raw.githubusercontent.com/laminlabs/lamindb/refs/heads/main/README.md"
    )
    artifact_readme = ln.Artifact(http_path, description="register http readme").save()
    # might change
    assert artifact_readme.hash is not None
    cache_path = artifact_readme.cache()
    assert cache_path.exists()
    assert cache_path.stat().st_size == http_path.stat().st_size
    cache_path.unlink()
    # just check saving for the second time (when Strage record is in the db)
    artifact_license = ln.Artifact(
        "https://raw.githubusercontent.com/laminlabs/lamindb/refs/heads/main/LICENSE",
        description="register http license",
    ).save()
    assert artifact_license.hash == "IQxRSNjvb7w2OLFeWqYlsg"

    artifact_readme.delete(permanent=True, storage=False)
    artifact_license.delete(permanent=True, storage=False)


# also see test in lamindb-setup/tests/storage/test_storage_stats.py
# there is also a test for GCP there
def test_folder_like_artifact_s3():
    study0_data = ln.Artifact("s3://lamindata/iris_studies/study0_raw_images")
    assert study0_data.hash == "IVKGMfNwi8zKvnpaD_gG7w"
    assert study0_data._hash_type == "md5-d"
    assert study0_data.n_files == 51
    assert study0_data.size == 658465



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.request import urlretrieve

import anndata as ad
import pandas as pd
from upath import UPath

from lamindb.base.ids import base62
from lamindb.core._settings import settings

if TYPE_CHECKING:
    from mudata import MuData
    from spatialdata import SpatialData


def file_fcs() -> Path:
    """Example FCS artifact."""
    filepath, _ = urlretrieve(
        "https://lamindb-dev-datasets.s3.amazonaws.com/.lamindb/DBNEczSgBui0bbzBXMGH.fcs",
        "example.fcs",
    )
    return Path(filepath)


def file_fcs_alpert19(populate_registries: bool = False) -> Path:  # pragma: no cover
    """FCS file from Alpert19.

    Args:
        populate_registries: pre-populate metadata records to simulate existing registries  # noqa
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/Alpert19-070314-Mike-Study+15-2013-plate+1-15-004-1-13_cells_found.fcs",
        "Alpert19.fcs",
    )
    if populate_registries:
        import bionty as bt
        import readfcs

        import lamindb as ln

        verbosity = ln.settings.verbosity
        ln.settings.verbosity = "error"
        adata = readfcs.read(filepath)
        std = bt.CellMarker.public().standardize(adata.var.index)
        ln.save(
            bt.CellMarker.from_values(
                bt.CellMarker.public().inspect(std, "name").validated, "name"
            )
        )
        ln.Feature(name="assay", dtype=[bt.ExperimentalFactor]).save()  # type: ignore
        ln.Feature(name="organism", dtype=[bt.Organism]).save()  # type: ignore
        ln.settings.verbosity = verbosity
    return Path(filepath)


def file_jpg_paradisi05() -> Path:
    """Return jpg file example.

    Originally from: https://upload.wikimedia.org/wikipedia/commons/2/28/Laminopathic_nuclei.jpg
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/Laminopathic_nuclei.jpg",
        "paradisi05_laminopathic_nuclei.jpg",
    )
    return Path(filepath)


def file_tsv_rnaseq_nfcore_salmon_merged_gene_counts(
    populate_registries: bool = False,
) -> Path:  # pragma: no cover
    """Gene counts table from nf-core RNA-seq pipeline.

    Output of: https://nf-co.re/rnaseq
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/salmon.merged.gene_counts.tsv",
        "salmon.merged.gene_counts.tsv",
    )
    if populate_registries:
        import bionty as bt

        import lamindb as ln

        verbosity = ln.settings.verbosity
        ln.settings.verbosity = "error"
        ln.Feature(name="assay", dtype=[bt.ExperimentalFactor]).save()  # type: ignore
        ln.Feature(name="organism", dtype=[bt.Organism]).save()  # type: ignore
        bt.ExperimentalFactor.from_source(ontology_id="EFO:0008896").save()
        ln.settings.verbosity = verbosity

    return Path(filepath)


def file_fastq(in_storage_root=False) -> Path:
    """Mini mock fastq artifact."""
    basedir = Path() if not in_storage_root else settings.storage.root
    filepath = basedir / "input.fastq.gz"
    with open(filepath, "w") as f:
        f.write("Mock fastq artifact.")
    return filepath


def file_bam(in_storage_root=False) -> Path:  # pragma: no cover
    """Mini mock bam artifact."""
    basedir = Path() if not in_storage_root else settings.storage.root
    filepath = basedir / "output.bam"
    with open(filepath, "w") as f:
        f.write("Mock bam artifact.")
    return filepath


def file_mini_csv(in_storage_root=False) -> Path:
    """Mini csv artifact."""
    basedir = Path() if not in_storage_root else settings.storage.root
    filepath = basedir / "mini.csv"
    df = pd.DataFrame([1, 2, 3], columns=["test"])
    df.to_csv(filepath, index=False)
    return filepath


def file_tiff_suo22() -> Path:  # pragma: no cover
    """Image file from Suo22.

    Pair with anndata_suo22_Visium10X
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/F121_LP1_4LIV.tiff",
        "F121_LP1_4LIV.tiff",
    )
    Path("suo22/").mkdir(exist_ok=True)
    filepath = Path(filepath).rename("suo22/F121_LP1_4LIV.tiff")  # type: ignore
    return Path(filepath)


def dir_iris_images() -> UPath:  # pragma: no cover
    """Directory with 3 studies of the Iris flower: 405 images & metadata.

    Provenance: https://lamin.ai/laminlabs/lamindata/transform/3q4MpQxRL2qZ5zKv

    The problem is that the same artifact was also ingested by the downstream
    demo notebook:
    https://lamin.ai/laminlabs/lamindata/transform/NJvdsWWbJlZS5zKv

    This is why on the UI, the artifact shows up as output of the downstream
    demo notebook rather than the upstream curation notebook. The lineage
    information should still be captured by
    https://github.com/laminlabs/lnschema-core/blob/a90437e91dfbd6b9002f18c3e978bd0f9c9a632d/lamindb/models.py#L2050-L2052
    but we don't use this in the UI yet.
    """
    return UPath("s3://lamindata/iris_studies")


def anndata_mouse_sc_lymph_node(
    populate_registries: bool = False,
) -> ad.AnnData:  # pragma: no cover
    """Mouse lymph node scRNA-seq collection from EBI.

    Subsampled to 10k genes.

    From: https://www.ebi.ac.uk/arrayexpress/experiments/E-MTAB-8414/

    Args:
        populate_registries: pre-populate metadata records to simulate existing registries  # noqa
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/E-MTAB-8414.h5ad")
    adata = ad.read_h5ad(filepath)

    # The column names are a bit lengthy, let's abbreviate them:
    adata.obs.columns = (
        adata.obs.columns.str.replace("Sample Characteristic", "")
        .str.replace("Factor Value ", "Factor Value:", regex=True)
        .str.replace("Factor Value\\[", "Factor Value:", regex=True)
        .str.replace(" Ontology Term\\[", "ontology_id:", regex=True)
        .str.strip("[]")
        .str.replace("organism part", "tissue")
        .str.replace("organism", "organism")
        .str.replace("developmental stage", "developmental_stage")
        .str.replace("cell type", "cell_type")
        # the last one could be interesting, too
        # .str.replace("Factor Value:Ontology Term[inferred cell_type - authors labels", "cell_type_authors")
    )
    # subset columns to only the ones with names
    columns = [
        col
        for col in adata.obs.columns
        if not col.startswith("ontology_id")
        and not col.startswith("Factor Value")
        and col != "strain"
    ]
    adata.obs = adata.obs[columns]

    # pre-populate registries
    if populate_registries:
        import bionty as bt

        import lamindb as ln

        verbosity = ln.settings.verbosity
        ln.settings.verbosity = "error"
        # strain
        bt.ExperimentalFactor.from_source(ontology_id="EFO:0004472").save()
        # developmental stage
        bt.ExperimentalFactor.from_source(ontology_id="EFO:0001272").save()
        # tissue
        bt.Tissue.from_source(ontology_id="UBERON:0001542").save()
        # cell types
        ln.save(bt.CellType.from_values(["CL:0000115", "CL:0000738"], "ontology_id"))
        # assays
        ln.Feature(name="assay", dtype=[bt.ExperimentalFactor]).save()  # type: ignore
        bt.ExperimentalFactor.from_source(ontology_id="EFO:0008913").save()
        # genes
        validated = bt.Gene.public(organism="mouse").validate(
            adata.var.index, field="ensembl_gene_id"
        )
        ln.save(
            bt.Gene.from_values(
                adata.var.index[validated][:-19],
                field="ensembl_gene_id",
                organism="mouse",
            )
        )
        # labels
        labels = []
        for col in ["sex", "age", "genotype", "immunophenotype"]:
            labels += [ln.ULabel(name=name) for name in adata.obs[col]]
        ln.save(labels)
        ln.settings.verbosity = verbosity

    return adata


def anndata_pbmc68k_reduced() -> ad.AnnData:
    """Modified from scanpy.collections.pbmc68k_reduced().

    This code was run::

        pbmc68k = sc.collections.pbmc68k_reduced()
        pbmc68k.obs.rename(columns={"bulk_labels": "cell_type"}, inplace=True)
        pbmc68k.obs["cell_type"] = pbmc68k.obs["cell_type"].cat.rename_categories(
            {"Dendritic": "Dendritic cells", "CD14+ Monocyte": "CD14+ Monocytes"}
        )
        del pbmc68k.obs["G2M_score"]
        del pbmc68k.obs["S_score"]
        del pbmc68k.obs["phase"]
        del pbmc68k.obs["n_counts"]
        del pbmc68k.var["dispersions"]
        del pbmc68k.var["dispersions_norm"]
        del pbmc68k.var["means"]
        del pbmc68k.uns["rank_genes_groups"]
        del pbmc68k.uns["bulk_labels_colors"]
        sc.pp.subsample(pbmc68k, fraction=0.1, random_state=123)
        pbmc68k.write("scrnaseq_pbmc68k_tiny.h5ad")
    """
    filepath, _ = urlretrieve(
        "https://lamindb-dev-datasets.s3.amazonaws.com/scrnaseq_pbmc68k_tiny.h5ad"
    )
    return ad.read_h5ad(filepath)


def anndata_file_pbmc68k_test() -> Path:
    """Modified from scanpy.collections.pbmc68k_reduced().

    Additional slots were added for testing purposes. Returns the filepath.

    To reproduce::

        pbmc68k = ln.examples.datasets.anndata_pbmc68k_reduced()
        pbmc68k_test = pbmc68k[:30, :200].copy()
        pbmc68k_test.raw = pbmc68k_test[:, :100]
        pbmc68k_test.obsp["test"] = sparse.eye(pbmc68k_test.shape[0], format="csr")
        pbmc68k_test.varp["test"] = sparse.eye(pbmc68k_test.shape[1], format="csr")
        pbmc68k_test.layers["test"] = sparse.csr_matrix(pbmc68k_test.shape)
        pbmc68k_test.layers["test"][0] = 1.
        pbmc68k_test.write("pbmc68k_test.h5ad")
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/pbmc68k_test.h5ad", "pbmc68k_test.h5ad"
    )
    return Path(filepath)


def anndata_pbmc3k_processed() -> ad.AnnData:  # pragma: no cover
    """Modified from scanpy.pbmc3k_processed()."""
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/scrnaseq_scanpy_pbmc3k_processed.h5ad"
    )
    pbmc3k = ad.read_h5ad(filepath)
    pbmc3k.obs.rename(columns={"louvain": "cell_type"}, inplace=True)
    return pbmc3k


def anndata_human_immune_cells(
    populate_registries: bool = False,
) -> ad.AnnData:  # pragma: no cover
    """Cross-tissue immune cell analysis reveals tissue-specific features in humans.

    From: https://cellxgene.cziscience.com/collections/62ef75e4-cbea-454e-a0ce-998ec40223d3
    Collection: Global

    To reproduce the subsample::
        >>> adata = sc.read('Global.h5ad')
        >>> adata.obs = adata.obs[['donor_id', 'tissue', 'cell_type', 'assay', 'tissue_ontology_term_id', 'cell_type_ontology_term_id', 'assay_ontology_term_id']].copy()
        >>> sc.pp.subsample(adata, fraction=0.005)
        >>> del adata.uns["development_cache_ontology_term_id_colors"]
        >>> del adata.uns["sex_ontology_term_id_colors"]
        >>> adata.write('human_immune.h5ad')
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/human_immune.h5ad")
    adata = ad.read_h5ad(filepath)
    adata.var.drop(columns=["gene_symbols", "feature_name"], inplace=True)
    adata.uns.pop("cell_type_ontology_term_id_colors")
    adata.uns.pop("title")
    adata.uns.pop("schema_version")
    adata.obs.columns = adata.obs.columns.str.replace("donor_id", "donor")
    columns = [col for col in adata.obs.columns if "ontology_term" not in col]
    adata.obs = adata.obs[columns]
    if populate_registries:
        import bionty as bt

        import lamindb as ln

        ln.save(
            bt.Gene.from_values(
                adata.var.index, field="ensembl_gene_id", organism="human"
            )
        )
        ln.save(bt.CellType.from_values(adata.obs.cell_type, field="name"))
        ln.save(bt.ExperimentalFactor.from_values(adata.obs.assay, field="name"))
        ln.save(bt.Tissue.from_values(adata.obs.tissue, field="name"))
        ln.Feature(name="cell_type", dtype=[bt.CellType]).save()  # type: ignore
        ln.Feature(name="assay", dtype=[bt.ExperimentalFactor]).save()  # type: ignore
        ln.Feature(name="tissue", dtype=[bt.Tissue]).save()  # type: ignore
        ln.Feature(name="organism", dtype=[bt.Organism]).save()  # type: ignore
        ln.Feature(name="donor", dtype=[ln.ULabel]).save()  # type: ignore
        bt.ExperimentalFactor.from_source(ontology_id="EFO:0008913").save()
        ln.save([ln.ULabel(name=name) for name in adata.obs.donor.unique()])
    return adata


def anndata_suo22_Visium10X():  # pragma: no cover
    """AnnData from Suo22 generated by 10x Visium."""
    import anndata as ad

    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/suo22_Visium10X_data_LI_subset.h5ad",
        "Visium10X_data_LI_subset.h5ad",
    )
    Path("suo22/").mkdir(exist_ok=True)
    filepath = Path(filepath).rename("suo22/Visium10X_data_LI_subset.h5ad")
    return ad.read_h5ad(filepath)


def mudata_papalexi21_subset(with_uns: bool = False) -> MuData:  # pragma: no cover
    """A subsetted mudata from papalexi21.

    To reproduce the subsetting:
        >>> !wget https://figshare.com/ndownloader/files/36509460
        >>> import mudata as md
        >>> import scanpy as sc
        >>> mdata = md.read_h5mu("36509460")
        >>> mdata = sc.pp.subsample(mdata, n_obs=200, copy=True)[0]
        >>> mdata[:, -300:].copy().write("papalexi21_subset_200x300_lamindb_demo_2023-07-25.h5mu")
    """
    import mudata as md

    md.set_options(pull_on_update=False)

    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/papalexi21_subset_200x300_lamindb_demo_2023-07-25.h5mu",
        "papalexi21_subset.h5mu",
    )

    mdata = md.read_h5mu(filepath)

    mdata.pull_obs()

    # The MuData object is malformed with duplicated information
    # Drop all columns for the modalities and add them again correspondingly
    for mod in ["rna", "adt", "hto", "gdo"]:
        mdata[mod].obs.drop(mdata[mod].obs.columns, axis=1, inplace=True)
    for col in mdata.obs.columns:
        for mod in ["rna", "adt", "hto", "gdo"]:
            if col.endswith(f"_{mod.upper()}"):
                new_col = col.replace(f"{mod}:", "")
                if new_col != col:
                    mdata[mod].obs[new_col] = mdata.obs.pop(col)
            else:
                new_col = col.replace(f"{mod}:", "")
                if new_col not in mdata.obs.columns and col in mdata.obs.columns:
                    mdata.obs[new_col] = mdata.obs.pop(col)

    for col in mdata.obs.columns:
        for mod in ["rna", "adt", "hto", "gdo"]:
            if col.endswith(f"_{mod.upper()}"):
                del mdata.obs[col]

    for col in [
        "orig.ident",
        "MULTI_ID",
        "NT",
        "S.Score",
        "G2M.Score",
        "Phase",
        "gene_target",
        "guide_ID",
        "HTO_classification",
    ]:
        del mdata.obs[col]

    mdata.push_obs(["percent.mito"], mods=["rna"], drop=True)
    mdata["hto"].obs["technique"] = "cell hashing"
    mdata["hto"].obs["technique"] = mdata["hto"].obs["technique"].astype("category")
    mdata.pull_obs(["technique"], mods="hto")

    if with_uns:
        mdata.uns["study_metadata"] = {
            "temperature": 21.6,
            "experiment": "Experiment 1",
        }
        mdata["rna"].uns["site_metadata"] = {"pos": 99.9, "site_id": "SITE001"}

    return mdata


def dict_cellxgene_uns() -> dict[str, Any]:
    """An example CELLxGENE AnnData `.uns` dictionary."""
    uns = {
        "organism_ontology_term_id": "NCBITaxon:9606",
        "spatial": {
            "is_single": True,
            "library_1": {  # Dynamic library_id key
                "images": {
                    "fullres": "path/to/fullres.jpg",
                    "hires": "path/to/hires.jpg",
                },
                "scalefactors": {
                    "spot_diameter_fullres": 89.43,
                    "tissue_hires_scalef": 0.177,
                },
            },
            "library_2": {  # Another dynamic library_id key
                "images": {
                    "fullres": "path/to/fullres_2.jpg",
                    "hires": "path/to/hires_2.jpg",
                },
                "scalefactors": {
                    "spot_diameter_fullres": 120.34,
                    "tissue_hires_scalef": 0.355,
                },
            },
        },
    }

    return uns


def df_iris() -> pd.DataFrame:
    """The iris collection as in sklearn.

    Original code::

        sklearn.collections.load_iris(as_frame=True).frame
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/iris.parquet")
    return pd.read_parquet(filepath)


def df_iris_in_meter() -> pd.DataFrame:
    """The iris collection with lengths in meter."""
    df = df_iris()
    # rename columns
    df.rename(
        columns={
            "sepal length (cm)": "sepal_length",
            "sepal width (cm)": "sepal_width",
            "petal length (cm)": "petal_length",
            "petal width (cm)": "petal_width",
        },
        inplace=True,
    )
    df[["sepal_length", "sepal_width", "petal_length", "petal_width"]] /= 100
    df["iris_organism_name"] = df["target"].map(
        {0: "setosa", 1: "versicolor", 2: "virginica"}
    )
    del df["target"]
    return df


def df_iris_in_meter_study1() -> pd.DataFrame:
    """The iris collection with lengths in meter."""
    df_iris = df_iris_in_meter()
    return df_iris.iloc[: len(df_iris) // 2]


def df_iris_in_meter_study2() -> pd.DataFrame:
    """The iris collection with lengths in meter."""
    df_iris = df_iris_in_meter()
    return df_iris.iloc[len(df_iris) // 2 :]


def dir_scrnaseq_cellranger(
    sample_name: str, basedir: str | Path = "./", output_only: bool = True
):  # pragma: no cover
    """Generate mock cell ranger outputs.

    Args:
        sample_name: name of the sample
        basedir: run directory
        output_only: only generate output files
    """
    basedir = Path(basedir)

    if not output_only:  # pragma: no cover
        fastqdir = basedir / "fastq"
        fastqdir.mkdir(parents=True, exist_ok=True)
        fastqfile1 = fastqdir / f"{sample_name}_R1_001.fastq.gz"
        with open(fastqfile1, "w") as f:
            f.write(f"{base62(n_char=6)}")
        fastqfile2 = fastqdir / f"{sample_name}_R2_001.fastq.gz"
        fastqfile2.touch(exist_ok=True)
        with open(fastqfile2, "w") as f:
            f.write(f"{base62(n_char=6)}")

    sampledir = basedir / f"{sample_name}"
    for folder in ["raw_feature_bc_matrix", "filtered_feature_bc_matrix", "analysis"]:
        filedir = sampledir / folder
        filedir.mkdir(parents=True, exist_ok=True)

    for filename in [
        "web_summary.html",
        "metrics_summary.csv",
        "possorted_genome_bam.bam",
        "possorted_genome_bam.bam.bai",
        "molecule_info.h5",
        "cloupe.cloupe",
        "raw_feature_bc_matrix.h5",
        "raw_feature_bc_matrix/barcodes.tsv.gz",
        "raw_feature_bc_matrix/features.tsv.gz",
        "raw_feature_bc_matrix/matrix.mtx.gz",
        "filtered_feature_bc_matrix.h5",
        "filtered_feature_bc_matrix/barcodes.tsv.gz",
        "filtered_feature_bc_matrix/features.tsv.gz",
        "filtered_feature_bc_matrix/matrix.mtx.gz",
        "analysis/analysis.csv",
    ]:
        file = sampledir / filename
        with open(file, "w") as f:
            f.write(f"{base62(n_char=6)}")

    return sampledir


def schmidt22_crispra_gws_IFNG(basedir=".") -> Path:  # pragma: no cover
    """CRISPRi screen collection of Schmidt22.

    Originally from: https://zenodo.org/record/5784651
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/schmidt22-crispra-gws-IFNG.csv",
        "schmidt22-crispra-gws-IFNG.csv",
    )
    return Path(filepath).rename(Path(basedir) / filepath)


def schmidt22_perturbseq(basedir=".") -> Path:  # pragma: no cover
    """Perturb-seq collection of Schmidt22.

    Subsampled and converted to h5ad from R file: https://zenodo.org/record/5784651

    To reproduce the subsample:
    >>> adata = sc.read('HuTcellsCRISPRaPerturbSeq_Re-stimulated.h5ad')
    >>> adata.obs = adata.obs[['cluster_name']]
    >>> del adata.obsp
    >>> del adata.var['features']
    >>> del adata.obsm['X_pca']
    >>> del adata.uns
    >>> del adata.raw
    >>> del adata.varm
    >>> adata.obs = adata.obs.reset_index()
    >>> del adata.obs['index']
    >>> sc.pp.subsample(adata, 0.03)
    >>> adata.write('schmidt22_perturbseq.h5ad')
    """
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/schmidt22_perturbseq.h5ad",
        "schmidt22_perturbseq.h5ad",
    )
    return Path(filepath).rename(Path(basedir) / filepath)


def spatialdata_blobs() -> SpatialData:
    """Example SpatialData dataset for tutorials."""
    from spatialdata.datasets import blobs

    sdata = blobs()
    sdata.attrs["bio"] = {
        "disease": "Alzheimer disease",
        "developmental_stage": "adult stage",
    }
    sdata.attrs["tech"] = {
        "assay": "Visium Spatial Gene Expression",
    }
    sdata.attrs["random_int"] = 20
    sdata.tables["table"].var.index = [
        "ENSG00000139618",  # BRCA2
        "ENSG00000157764",  # BRAF
        "ENSG00000999999",  # Does not exist
    ]
    sdata.tables["table"].obs["sample_region"] = pd.Categorical(
        ["sample region 1"] * 13 + ["sample region 2"] * 13
    )

    return sdata



from __future__ import annotations

import json
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import lamindb_setup as ln_setup
from lamin_utils import logger

from lamindb.models.artifact import Artifact
from lamindb.models.collection import Collection
from lamindb.models.run import Run
from lamindb.models.transform import Transform

if TYPE_CHECKING:
    from vitessce import VitessceConfig


# "unit test": https://github.com/laminlabs/lamindb/blob/main/docs/storage/vitessce.ipynb
# integration test & context: https://github.com/laminlabs/lamin-spatial/blob/main/docs/vitessce.ipynb
def save_vitessce_config(
    vitessce_config: VitessceConfig,
    key: str | None = None,
    description: str | None = None,
) -> Artifact:
    """Validates and saves a `VitessceConfig` object.

    If the `VitessceConfig` object references multiple artifacts, automatically
    creates a `Collection` and displays the "Vitessce button" next to it.

    The `VitessceConfig` artifact has `.suffix = ".vitessce.json"` and `.kind = "__lamindb_config__"`,
    which is by default hidden on the hub UI.

    Guide: :doc:`docs:vitessce`.

    Args:
        vitessce_config: A `VitessceConfig` object.
        key: A `key` for the `VitessceConfig` artifact.
        description: A `description` for the `VitessceConfig` aritifact. Is additionally
            used as `key` for a `Collection` in case the `VitessceConfig` object
            references multiple artifacts.
    """
    # can only import here because vitessce is not a dependency
    from vitessce import VitessceConfig

    assert isinstance(vitessce_config, VitessceConfig)  # noqa: S101
    vc_dict = vitessce_config.to_dict()
    try:
        url_to_artifact_dict = vitessce_config.get_artifacts()
    except AttributeError as e:
        raise SystemExit(
            "save_vitessce_config() requires vitessce>=3.4.0: pip install vitessce>=3.4.0"
        ) from e
    dataset_artifacts = list(url_to_artifact_dict.values())
    message = "\n".join([artifact.__repr__() for artifact in dataset_artifacts])
    logger.important(f"VitessceConfig references these artifacts:\n{message}")
    assert len(dataset_artifacts) > 0  # noqa: S101

    # the below will be replaced with a `ln.tracked()` decorator soon
    transform = Transform(  # type: ignore
        uid="kup03MJBsIVa0002",
        key="save_vitessce_config",
        type="function",
        version="3",
    ).save()
    run = Run(transform=transform).save()
    run.input_artifacts.set(dataset_artifacts)
    collection = None
    if len(dataset_artifacts) > 1:
        # if we have more datasets, we should create a collection
        # and attach an action to the collection
        # consicious use of description for key, see here
        # https://github.com/laminlabs/lamindb/pull/2997
        collection = Collection(dataset_artifacts, key=description).save()

    # create a JSON export
    config_file_local_path = ln_setup.settings.cache_dir / "config.vitessce.json"
    with open(config_file_local_path, "w") as file:
        json.dump(vc_dict, file)
    vitessce_config_artifact = Artifact(
        config_file_local_path,
        key=key,
        description=description,
        run=run,
        kind="__lamindb_config__",
    ).save()
    slug = ln_setup.settings.instance.slug
    logger.important(
        f"VitessceConfig: https://lamin.ai/{slug}/artifact/{vitessce_config_artifact.uid}"
    )
    if collection is None:
        # we have one and only one dataset artifact, hence the following line is OK
        dataset_artifacts[0]._actions.add(vitessce_config_artifact)
        logger.important(
            f"Dataset: https://lamin.ai/{slug}/artifact/{dataset_artifacts[0].uid}"
        )
    else:
        collection._actions.add(vitessce_config_artifact)
        logger.important(
            f"Collection: https://lamin.ai/{slug}/collection/{collection.uid}"
        )
    run.finished_at = datetime.now(timezone.utc)
    run.save()
    return vitessce_config_artifact



"""Integrations.

.. autosummary::
   :toctree: .

   save_vitessce_config
   save_tiledbsoma_experiment
   curate_from_croissant
   lightning
"""

from lamindb.core.storage import save_tiledbsoma_experiment

from ._croissant import curate_from_croissant
from ._vitessce import save_vitessce_config

__all__ = [
    "lightning",
    "save_tiledbsoma_experiment",
    "curate_from_croissant",
    "save_vitessce_config",
]



# Generated by Django 5.2 on 2025-07-26 15:55

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0115_alter_space_uid"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="artifact",
            name="unique_artifact_storage_key_hash",
        ),
        migrations.AlterField(
            model_name="record",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="text",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="url",
            field=lamindb.base.fields.URLField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="name",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=150, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="artifact",
            constraint=models.UniqueConstraint(
                fields=("storage", "key", "hash"),
                name="unique_artifact_storage_key_hash",
            ),
        ),
    ]



import shutil

import lamindb as ln
import pytest


@pytest.mark.parametrize("filepath_prefix", [None, "testdb/"])
def test_curate_artifact_from_croissant(filepath_prefix: str | None):
    croissant_path, dataset1_path = ln.examples.croissant.mini_immuno(
        n_files=1, filepath_prefix=filepath_prefix
    )
    artifact1 = ln.integrations.curate_from_croissant(croissant_path)
    assert (
        artifact1.description
        == "Mini immuno dataset - A few samples from the immunology dataset"
    )
    assert artifact1.key == "mini_immuno.anndata.zarr"
    assert artifact1.version == "1.0"
    assert (
        artifact1._key_is_virtual
        if filepath_prefix is None
        else not artifact1._key_is_virtual
    )
    license_label = artifact1.ulabels.get(
        name="https://creativecommons.org/licenses/by/4.0/"
    )
    project_label = artifact1.projects.get(name="Mini Immuno Project")

    # now mutate the dataset and create a new version
    croissant_path, dataset1_path = ln.examples.croissant.mini_immuno(
        n_files=1, filepath_prefix=filepath_prefix, strip_version=True
    )
    dummy_file_path = dataset1_path / "dummy_file.txt"
    dummy_file_path.write_text("dummy file")

    artifact2 = ln.integrations.curate_from_croissant(croissant_path)
    assert artifact2.description == artifact1.description
    assert artifact2.key == artifact1.key
    assert artifact2.version is None
    assert artifact2.stem_uid == artifact1.stem_uid
    assert artifact2.uid != artifact1.uid
    assert (
        artifact2._key_is_virtual
        if filepath_prefix is None
        else not artifact1._key_is_virtual
    )
    license_label = artifact2.ulabels.get(
        name="https://creativecommons.org/licenses/by/4.0/"
    )
    project_label = artifact2.projects.get(name="Mini Immuno Project")

    shutil.rmtree(dataset1_path)
    croissant_path.unlink()
    artifact1.delete(permanent=True, storage=True)  # because of real storage key
    project_label.delete(permanent=True)
    license_label.delete(permanent=True)


def test_curate_collection_from_croissant():
    croissant_path, dataset1_path, dataset2_path = ln.examples.croissant.mini_immuno(
        n_files=2
    )
    collection = ln.integrations.curate_from_croissant(croissant_path)
    croissant_path.unlink()
    shutil.rmtree(dataset1_path)
    dataset2_path.unlink()
    artifact1 = collection.artifacts.get(key="mini_immuno.anndata.zarr")
    artifact2 = collection.artifacts.get(key="mini.csv")
    license_label = collection.ulabels.get(
        name="https://creativecommons.org/licenses/by/4.0/"
    )
    project_label = collection.projects.get(name="Mini Immuno Project")

    collection.delete(permanent=True)
    artifact1.delete(permanent=True)
    artifact2.delete(permanent=True)
    project_label.delete(permanent=True)
    license_label.delete(permanent=True)



# Generated by Django 5.1.12 on 2025-09-28 23:37

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "lamindb",
            "0123_alter_artifact_description_alter_branch_description_and_more",
        ),
    ]

    # this migration was later reverted, so we do not need to run it
    operations = []  # type: ignore



from __future__ import annotations

import builtins
import importlib
import inspect
from typing import TYPE_CHECKING

from lamin_utils import colors, logger
from lamindb_setup import settings
from lamindb_setup._init_instance import get_schema_module_name

from lamindb.models import Feature, FeatureValue, ParamValue, SQLRecord

from .models.feature import serialize_pandas_dtype

if TYPE_CHECKING:
    import pandas as pd

is_run_from_ipython = getattr(builtins, "__IPYTHON__", False)


def display_df_with_descriptions(
    df: pd.DataFrame, descriptions: dict[str, str] | None = None
):
    from IPython.display import HTML, display

    if descriptions is None:
        display(df)
        return None

    # Start building HTML table
    html = '<table class="dataframe">'

    # Create header with title and description rows
    html += "<thead>"

    # Column names row
    html += "<tr>"
    html += '<th class="header-title index-header"></th>'  # Index header
    for col in df.columns:
        html += f'<th class="header-title">{col}</th>'
    html += "</tr>"

    # Descriptions row
    html += "<tr>"
    html += f'<th class="header-desc index-header">{df.index.name or ""}</th>'  # Index column
    for col in df.columns:
        desc = descriptions.get(col, "")
        html += f'<th class="header-desc">{desc}</th>'
    html += "</tr>"

    html += "</thead>"

    # Add body rows
    html += "<tbody>"
    for idx, row in df.iterrows():
        html += "<tr>"
        html += f'<th class="row-index">{idx}</th>'  # Index value
        for col in df.columns:
            html += f"<td>{row[col]}</td>"
        html += "</tr>"
    html += "</tbody>"
    html += "</table>"

    # Add CSS styles
    styled_html = f"""
    <style>
        .dataframe {{
            border-collapse: collapse;
            margin: 10px 0;
        }}
        .dataframe th, .dataframe td {{
            border: 1px solid #ddd;
            padding: 8px;
            text-align: left;
        }}
        .header-title {{
            font-weight: bold;
        }}
        .header-desc {{
            color: #666;
            font-weight: normal;
        }}
        .row-index {{
            font-weight: bold;
        }}
        .index-header {{
            font-weight: bold;
        }}
    </style>
    {html}
    """
    return display(HTML(styled_html))


def view(
    df: pd.DataFrame | None = None,
    limit: int = 7,
    modules: str | None = None,
    registries: list[str] | None = None,
) -> None:
    """View metadata.

    Args:
        df: A DataFrame to display.
        limit: Display the latest `n` records
        modules: schema module to view. Default's to
            `None` and displays all registry modules.
        registries: List of SQLRecord names. Defaults to
            `None` and lists all registries.

    Examples:
        >>> ln.view()
    """
    if df is not None:
        descriptions = {
            col_name: serialize_pandas_dtype(dtype)
            for col_name, dtype in df.dtypes.to_dict().items()
        }
        feature_dtypes = dict(Feature.objects.values_list("name", "dtype"))
        descriptions.update(feature_dtypes)
        display_df_with_descriptions(df, descriptions)
        return None

    if is_run_from_ipython:
        from IPython.display import display as show
    else:
        show = logger.print

    if modules is not None:
        module_names = [modules]
    else:
        module_names = ["core"] + list(settings.instance.modules)

    for module_name in module_names:
        schema_module = importlib.import_module(get_schema_module_name(module_name))
        # the below is necessary because a schema module might not have been
        # explicitly accessed
        importlib.reload(schema_module)

        all_registries = {
            registry
            for registry in schema_module.__dict__.values()
            if inspect.isclass(registry)
            and issubclass(registry, SQLRecord)
            and registry is not SQLRecord
        }
        if module_name == "core":
            all_registries.update({FeatureValue, ParamValue})
        if registries is not None:
            filtered_registries = {
                registry
                for registry in all_registries
                if registry.__name__ in registries
            }
        else:
            filtered_registries = all_registries
        if len(module_names) > 1:
            section = f"* module: {colors.green(colors.bold(module_name))} *"
            section_no_color = f"* module: {module_name} *"
            logger.print("*" * len(section_no_color))
            logger.print(section)
            logger.print("*" * len(section_no_color))
        for registry in sorted(filtered_registries, key=lambda x: x.__name__):
            df = registry.to_dataframe(limit=limit)
            if df.shape[0] > 0:
                logger.print(colors.blue(colors.bold(registry.__name__)))
                show(df)



user_id_cache = {}


def current_user_id() -> int:
    import lamindb_setup as ln_setup
    from lamindb_setup import settings
    from lamindb_setup._init_instance import register_user

    from lamindb.models import User

    def query_user_id():
        if ln_setup.core.django.IS_MIGRATING:
            return 1
        else:
            try:
                user_id = User.objects.get(uid=settings.user.uid).id
            except User.DoesNotExist:
                register_user(settings.user)
                user_id = User.objects.get(uid=settings.user.uid).id
            return user_id

    if settings._instance_exists:
        if settings.instance.slug not in user_id_cache:
            user_id_cache[settings.instance.slug] = query_user_id()
        return user_id_cache[settings.instance.slug]
    else:
        return query_user_id()



# Generated by Django 5.1.12 on 2025-10-04 10:57

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0131_record_unique_name_type_space"),
    ]

    operations = [
        migrations.AddField(
            model_name="record",
            name="parents",
            field=models.ManyToManyField(related_name="children", to="lamindb.record"),
        ),
        migrations.AddField(
            model_name="record",
            name="reference",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AddField(
            model_name="record",
            name="reference_type",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=25, null=True
            ),
        ),
        migrations.RenameField(
            model_name="record",
            old_name="ulabels",
            new_name="linked_ulabels",
        ),
        migrations.AlterField(
            model_name="record",
            name="linked_ulabels",
            field=models.ManyToManyField(
                related_name="linked_in_records",
                through="lamindb.RecordULabel",
                to="lamindb.ulabel",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.run.current_run,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_records",
                to="lamindb.run",
            ),
        ),
    ]



import lamindb as ln


def test_manager_list():
    label = ln.Record(name="manager label")
    label.save()
    label_names = [f"Record {i}" for i in range(3)]
    labels = [ln.Record(name=name) for name in label_names]
    ln.save(labels)
    label.parents.set(labels)
    assert len(label.parents.to_list()) == 3
    assert "Record 1" in label.parents.to_list("name")
    label.delete(permanent=True)
    for label in labels:
        label.delete(permanent=True)
